"""
This module provides a `PEP-249`_ compliant DBAPI interface, for a websocket based
database driver (see also `exasol-websocket-api`_).

.. _PEP-249: https://peps.python.org/pep-0249/#interfaceerror
.. _exasol-websocket-api: https://github.com/exasol/websocket-api
"""

from exasol.driver.websocket._connection import Connection as DefaultConnection

# Re-export types and definitions required by dbapi2
from exasol.driver.websocket._errors import (
    DatabaseError,
    DataError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    ProgrammingError,
    Warning,
)
from exasol.driver.websocket._protocols import (
    Connection,
    Cursor,
)
from exasol.driver.websocket._types import (
    BINARY,
    DATETIME,
    NUMBER,
    ROWID,
    STRING,
    Date,
    DateFromTicks,
    Time,
    TimeFromTicks,
    Timestamp,
    TimestampFromTicks,
    TypeCode,
)

# Add remaining definitions

apilevel = "2.0"  # Required by spec. pylint: disable=C0103
threadsafety = 1  # Required by spec. pylint: disable=C0103
paramstyle = "qmark"  # Required by spec. pylint: disable=C0103


def connect(connection_class=DefaultConnection, **kwargs) -> Connection:
    """
    Creates a connection to the database.

    Args:
        connection_class: which shall be used to construct a connection object.
        kwargs: compatible with the provided connection_class.

    Returns:

        returns a dbapi2 complaint connection object.
    """
    connection = connection_class(**kwargs)
    return connection.connect()


__all__ = [
    # ----- Constants -----
    "apilevel",
    "threadsafety",
    "paramstyle",
    # ----- Errors -----
    "Warning",
    "Error",
    "InterfaceError",
    "DatabaseError",
    "DataError",
    "OperationalError",
    "InternalError",
    "IntegrityError",
    "InternalError",
    "ProgrammingError",
    "NotSupportedError",
    # ----- Protocols -----
    "Connection",
    "Cursor",
    # ----- Types and Type-Conversions -----
    "Date",
    "Time",
    "Timestamp",
    "DateFromTicks",
    "TimeFromTicks",
    "TimestampFromTicks",
    "STRING",
    "BINARY",
    "NUMBER",
    "DATETIME",
    "ROWID",
    # ----- Functions ------
    "connect",
    # ----- Non DBAPI exports -----
    "TypeCode",
]



"""
This module provides `PEP-249`_ DBAPI compliant connection implementation.
(see also `PEP-249-connection`_)

.. _PEP-249-connection: https://peps.python.org/pep-0249/#connection-objects
"""

import ssl
from functools import wraps
from typing import Optional

import pyexasol
from exasol.driver.websocket._cursor import Cursor as DefaultCursor
from exasol.driver.websocket._errors import Error


def _requires_connection(method):
    """
    Decorator requires the object to have a working connection.

    Raises:
        Error if the connection object has no active connection.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self._connection:
            raise Error("No active connection available")
        return method(self, *args, **kwargs)

    return wrapper


class Connection:
    """
    Implementation of a websocket-based connection.

    For more details see :class: `Connection` protocol definition.
    """

    def __init__(
        self,
        dsn: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        schema: str = "",
        autocommit: bool = True,
        tls: bool = True,
        certificate_validation: bool = True,
        client_name: str = "EXASOL:DBAPI2:WS",
        client_version: str = "unknown",
    ):
        """
        Create a Connection object.

        Args:

            dsn: Connection string, same format as for standard JDBC / ODBC drivers.
            username: which will be used for the authentication.
            password: which will be used for the authentication.
            schema: to open after connecting.
            autocommit: enable autocommit.
            tls: enable tls.
            certificate_validation: disable certificate validation.
            client_name: which is communicated to the DB server.
        """

        # for more details see pyexasol.connection.ExaConnection
        self._options = {
            "dsn": dsn,
            "user": username,
            "password": password,
            "schema": schema,
            "autocommit": autocommit,
            "snapshot_transactions": None,
            "connection_timeout": 10,
            "socket_timeout": 30,
            "query_timeout": 0,
            "compression": False,
            "encryption": tls,
            "fetch_dict": False,
            "fetch_mapper": None,
            "fetch_size_bytes": 5 * 1024 * 1024,
            "lower_ident": False,
            "quote_ident": False,
            "json_lib": "json",
            "verbose_error": True,
            "debug": False,
            "debug_logdir": None,
            "udf_output_bind_address": None,
            "udf_output_connect_address": None,
            "udf_output_dir": None,
            "http_proxy": None,
            "client_name": client_name,
            "client_version": client_version,
            "protocol_version": 3,
            "websocket_sslopt": (
                {"cert_reqs": ssl.CERT_REQUIRED}
                if certificate_validation
                else {"cert_reqs": ssl.CERT_NONE}
            ),
            "access_token": None,
            "refresh_token": None,
        }
        self._connection = None

    def connect(self):
        """See also :py:meth: `Connection.connect`"""
        try:
            self._connection = pyexasol.connect(**self._options)
        except pyexasol.exceptions.ExaConnectionError as ex:
            raise Error(f"Connection failed, {ex}") from ex
        except Exception as ex:
            raise Error() from ex
        return self

    @property
    def connection(self):
        """Underlying connection used by this Connection"""
        return self._connection

    def close(self):
        """See also :py:meth: `Connection.close`"""
        connection_to_close = self._connection
        self._connection = None
        if connection_to_close is None or connection_to_close.is_closed:
            return
        try:
            connection_to_close.close()
        except Exception as ex:
            raise Error() from ex

    @_requires_connection
    def commit(self):
        """See also :py:meth: `Connection.commit`"""
        try:
            self._connection.commit()
        except Exception as ex:
            raise Error() from ex

    @_requires_connection
    def rollback(self):
        """See also :py:meth: `Connection.rollback`"""
        try:
            self._connection.rollback()
        except Exception as ex:
            raise Error() from ex

    @_requires_connection
    def cursor(self):
        """See also :py:meth: `Connection.cursor`"""
        return DefaultCursor(self)

    def __del__(self):
        if self._connection is None:
            return

        # Currently, the only way to handle this gracefully is to invoke the`__del__`
        # method of the underlying connection rather than calling an explicit `close`.
        #
        # For more details, see also:
        # * https://github.com/exasol/sqlalchemy-exasol/issues/390
        # * https://github.com/exasol/pyexasol/issues/108
        #
        # If the above tickets are resolved, it should be safe to switch back to using
        # `close` instead of `__del__`.
        self._connection.__del__()



import base64
import getpass
import hashlib
import itertools
import platform
import random
import re
import socket
import ssl
import threading
import time
import urllib.parse
import zlib
from collections.abc import Iterable
from inspect import (
    Signature,
    cleandoc,
    signature,
)
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Callable,
    NamedTuple,
    Optional,
    Union,
)
from warnings import warn

import rsa
import websocket
from packaging.version import Version

from . import callback as cb
from .exceptions import *
from .ext import ExaExtension
from .formatter import ExaFormatter
from .http_transport import (
    ExaHttpThread,
    ExaSQLExportThread,
    ExaSQLImportThread,
)
from .logger import ExaLogger
from .meta import ExaMetaData
from .script_output import ExaScriptOutputProcess
from .statement import ExaStatement
from .version import __version__
from .warnings import PyexasolWarning

if TYPE_CHECKING:
    import pandas
    import polars


class Host(NamedTuple):
    """This represents a resolved host name with its IP address and port number."""

    hostname: str
    ip_address: Optional[str]
    port: int
    fingerprint: Optional[str]


def get_exaconnection_signature() -> Signature:
    return signature(ExaConnection.__init__)


class ExaConnection:
    """
    Warning:
        Threads may share the module, but not connections
        One connection may be used by different threads, just not at the same time
        :meth:`pyexasol.ExaConnection.abort_query` is an exception,
        it is meant to be called from another thread

    Note:

        It is advisable to use multiprocessing instead of threading and create
        a new connection in each sub-process

        Public Attributes:
            ``attr``:
                Read-only ``dict`` of attributes of current connection.

            ``login_info``:
                Read-only ``dict`` of login information returned by second
                response of LOGIN command.

            ``options``:
                Read-only ``dict`` of arguments passed to
                :meth:`pyexasol.ExaConnection.connect`.
    """

    cls_statement = ExaStatement
    cls_formatter = ExaFormatter
    cls_logger = ExaLogger
    cls_extension = ExaExtension
    cls_meta = ExaMetaData

    threadsafety = 1

    def __init__(
        self,
        dsn: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        schema: str = "",
        autocommit=constant.DEFAULT_AUTOCOMMIT,
        snapshot_transactions=None,
        connection_timeout=constant.DEFAULT_CONNECTION_TIMEOUT,
        socket_timeout=constant.DEFAULT_SOCKET_TIMEOUT,
        query_timeout=constant.DEFAULT_QUERY_TIMEOUT,
        compression: bool = False,
        encryption: bool = True,
        fetch_dict: bool = False,
        fetch_mapper=None,
        fetch_size_bytes=constant.DEFAULT_FETCH_SIZE_BYTES,
        lower_ident: bool = False,
        quote_ident: bool = False,
        json_lib: str = "json",
        verbose_error: bool = True,
        debug: bool = False,
        debug_logdir=None,
        udf_output_bind_address=None,
        udf_output_connect_address=None,
        udf_output_dir=None,
        http_proxy=None,
        resolve_hostnames: bool = True,
        client_name=None,
        client_version=None,
        client_os_username=None,
        protocol_version=constant.PROTOCOL_V3,
        websocket_sslopt: Optional[dict] = None,
        access_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
    ):
        """
        Exasol connection object

        Args:
            dsn:
                Connection string, same format as standard JDBC / ODBC drivers
                (e.g. 10.10.127.1..11:8564)
            user:
                Username
            password:
                Password
            schema:
                Open schema after connection
                (Default: '', no schema)
            autocommit:
                Enable autocommit on connection
                (Default: True)
            snapshot_transactions:
                Explicitly enable or disable snapshot transactions on connection
                (Default: None, database default)
            connection_timeout:
                Socket timeout in seconds used to establish connection
                (Default: 10)
            socket_timeout:
                Socket timeout in seconds used for requests after connection was established
                (Default: 30)
            query_timeout:
                Maximum execution time of queries before automatic abort, in seconds
                (Default: 0, no timeout)
            compression:
                Use zlib compression both for WebSocket and HTTP transport
                (Default: False)
            encryption:
                Use SSL to encrypt client-server communications for WebSocket and HTTP transport
                (Default: True)
            fetch_dict:
                Fetch result rows as dicts instead of tuples (Default: False)
            fetch_mapper:
                Use custom mapper function to convert Exasol values into
                Python objects during fetching
                (Default: None)
            fetch_size_bytes:
                Maximum size of data message for single fetch request in bytes
                (Default: 5Mb)
            lower_ident:
                Automatically lowercase identifiers (table names, column names, etc.)
                returned from relevant functions
                (Default: False)
            quote_ident:
                Add double quotes and escape identifiers passed to relevant functions
                (export_*, import_*, ext.*, etc.)
                (Default: False)
            json_lib:
                Supported values: rapidjson, ujson, orjson, json
                (Default: json)
            verbose_error:
                Display additional information when error occurs
                (Default: True)
            debug:
                Output debug information for client-server communication and
                connection attempts to STDERR
            debug_logdir:
                Store debug information into files in debug_logdir instead of
                outputting it to STDERR
            udf_output_bind_address:
                Specific server_address to bind TCP server for UDF script output
                (default: ('', 0))
            udf_output_connect_address:
                Specific SCRIPT_OUTPUT_ADDRESS value to connect from Exasol to
                UDF script output server
                (default: inherited from TCP server)
            udf_output_dir:
                Directory to store captured UDF script output logs, split by
                <session_id>_<statement_id>/<vm_num>
            http_proxy:
                HTTP proxy string in Linux http_proxy format
                (default: None)
            resolve_hostnames:
                Explicitly resolve host names to IP addresses before connecting.
                Deactivating this will let the operating system resolve the host name
                (default: True)
            client_name:
                Custom name of client application displayed in Exasol sessions tables
                (Default: PyExasol)
            client_version:
                Custom version of client application
                (Default: pyexasol.__version__)
            client_os_username:
                Custom OS username displayed in Exasol sessions table
                (Default: getpass.getuser())
            protocol_version:
                Major WebSocket protocol version requested for connection
                (Default: pyexasol.PROTOCOL_V3)
            websocket_sslopt:
                Set custom SSL options for WebSocket client
                (Default: None)
            access_token:
                OpenID access token to use for the login process
            refresh_token:
                OpenID refresh token to use for the login process
        """

        # convert all arguments to a dict[argument_name, argument_value]
        sig = get_exaconnection_signature()
        all_locals = locals()
        self.options = {
            param.name: all_locals[param.name]
            for param in sig.parameters.values()
            if param.name != "self"
        }

        self.login_info: dict = {}
        self.login_time = 0
        self.attr: dict = {}
        self.is_closed: bool = False

        self.ws_req_count = 0
        self.ws_req_time = 0

        self.last_stmt = None
        self.stmt_count = 0

        self.json_encode = None
        self.json_decode = None

        self._udf_output_count = 0
        self._req_lock = threading.Lock()

        self._init_format()
        self._init_json()
        self._init_ext()
        self._init_meta()

        self._init_logger()
        self._init_ws()

        self._login()
        self.get_attr()

    def execute(self, query: str, query_params: Optional[dict] = None) -> ExaStatement:
        """
        Execute SQL query with optional query formatting parameters

        Args:
            query:
                SQL query text, possibly with placeholders
            query_params:
                 Values for placeholders

        Returns:
            ExaStatement object

        Examples:

            >>> con = ExaConnection(...)
            >>> con.execute(
            ...        query="SELECT * FROM {table!i} WHERE col1={col1}",
            ...        query_params={'table': 'users', 'col1':'bar'}
            ...)
        """
        return self.cls_statement(self, query, query_params)

    def execute_udf_output(self, query: str, query_params: Optional[dict] = None):
        """
        Execute SQL query with UDF script, capture output

        Note:
            Exasol should be able to open connection to the machine where current script is running.
            It is usually OK in the same data centre, but it is normally not working
            if you try to run this function on local laptop.

        Args:
            query:
                SQL query text, possibly with placeholders
            query_params:
                Values for placeholders

        Returns:
            Return tuple with two elements: (1) instance of :class:`pyexasol.ExaStatement`
            and (2) list of :class:`Path` objects for script output log files.

        Attention:
            Exasol should be able to open connection to the machine where current script is running

        Examples:
            >>> con = ExaConnection(...)
            >>> stmt, output_files = con.execute_udf_output(
            ...        query="SELECT * FROM {table!i} WHERE col1={col1}",
            ...        query_params={'table': 'users', 'col1':'bar'}
            ...)
        """
        stmt_output_dir = self._get_stmt_output_dir()

        script_output = ExaScriptOutputProcess(
            (
                self.options["udf_output_bind_address"][0]
                if self.options["udf_output_bind_address"]
                else None
            ),
            (
                self.options["udf_output_bind_address"][1]
                if self.options["udf_output_bind_address"]
                else None
            ),
            stmt_output_dir,
        )

        try:
            script_output.start()

            # This option is useful to get around complex network setups, like Exasol running in Docker containers
            if self.options["udf_output_connect_address"]:
                address = f"{self.options['udf_output_connect_address'][0]}:{self.options['udf_output_connect_address'][1]}"
            else:
                address = script_output.get_output_address()

            self.execute(
                "ALTER SESSION SET SCRIPT_OUTPUT_ADDRESS = {address}",
                {"address": address},
            )

            stmt = self.execute(query, query_params)
            log_files = sorted(list(stmt_output_dir.glob("*.log")))

            if len(log_files) > 0:
                script_output.join_with_exc()
            else:
                # In some cases Exasol does not run any VM's even when UDF scripts are being called
                # In this case we must terminate TCP server, since it won't stop automatically
                script_output.terminate()
                script_output.join()
        except ExaQueryError:
            script_output.terminate()
            script_output.join()

            raise

        return stmt, log_files

    def commit(self):
        """Wrapper for query 'COMMIT'"""
        return self.execute("COMMIT")

    def rollback(self):
        """Wrapper for query 'ROLLBACK'"""
        return self.execute("ROLLBACK")

    def set_autocommit(self, val: str) -> None:
        """
        Set autocommit mode.

        Args:
            val:
                Set ``False`` to execute following statements in transaction.
                Set ``True`` to get back to automatic COMMIT after each statement.

        Note:
            Autocommit is ``True`` by default because Exasol has to commit indexes and statistics
            objects even for pure SELECT statements. Lack of default COMMIT may lead to serious
            performance degradation.
        """
        if not isinstance(val, bool):
            raise ValueError("Autocommit value must be boolean")

        self.set_attr({"autocommit": val})

    def set_query_timeout(self, val):
        """
        Set the maximum time in seconds for which a query can run before Exasol kills it automatically.

        Args:
            val:
                Timeout value in seconds.
                Set value ``0`` to disable timeout.

        Note:
            It is highly recommended to set timeout for UDF scripts to
            avoid potential infinite loops and very long transactions.
        """
        self.set_attr({"queryTimeout": int(val)})

    def open_schema(self, schema):
        """
        Wrapper for `OPEN SCHEMA`

        Args:
            schema: Schema name
        """
        self.set_attr({"currentSchema": self.format.default_format_ident_value(schema)})

    def current_schema(self):
        """
        Get the name of the current schema.

        Returns:
            Name of currently opened schema. Returns an empty string if no schema was opened.
        """
        return self.attr.get("currentSchema", "")

    def export_to_file(
        self,
        dst,
        query_or_table: str,
        query_params: Optional[dict] = None,
        export_params: Optional[dict] = None,
    ):
        """
        Export large amount of data from Exasol to file or file-like object using fast HTTP transport.

        Note:
            File must be opened in binary mode.

        Args:
            dst:
                Path to file or file-like object where data will be exported to.
            query_or_table:
                SQL query or table from which to export data.
            query_params:
                Values for SQL query placeholders.
            export_params:
                Custom parameters for EXPORT query.

        Examples:
            >>> con = ExaConnection(...)
            >>> with open('/tmp/file.csv', 'wb') as f:
            ...     con.export_to_file(
            ...         dst=f,
            ...         query_or_table="SELECT * FROM table"
            ...     )
        """
        return self.export_to_callback(
            cb.export_to_file, dst, query_or_table, query_params, None, export_params
        )

    def export_to_list(
        self,
        query_or_table: str,
        query_params: Optional[dict] = None,
        export_params: Optional[dict] = None,
    ) -> list:
        """
        Export large amount of data from Exasol to basic Python `list` using fast HTTP transport.

        Args:
            query_or_table:
                SQL query or table from which to export data.
            query_params:
                Values for SQL query placeholders.
            export_params:
                Custom parameters for EXPORT query.

        Returns:
            `list` of `tuples`

        Warnings:
            - This function may run out of memory

        Examples:
            >>> con = ExaConnection(...)
            >>> myresult = con.export_to_list(
            ...    query_or_table="SELECT * FROM table"
            ... )
        """
        return self.export_to_callback(
            cb.export_to_list, None, query_or_table, query_params, None, export_params
        )

    def export_to_pandas(
        self,
        query_or_table: str,
        query_params: Optional[dict] = None,
        callback_params: Optional[dict] = None,
        export_params: Optional[dict] = None,
    ) -> "pandas.DataFrame":
        """
        Export large amount of data from Exasol to :class:`pandas.DataFrame`.

        Args:
            query_or_table:
                SQL query or table from which to export data.
            query_params:
                Values for SQL query placeholders.
            callback_params:
                Dictionary with additional parameters for callback function
                `pandas.read_csv <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html>`__.
            export_params:
                 Custom parameters for EXPORT query.

        Returns:
            instance of :class:`pandas.DataFrame`

        Warnings:
            - This function may run out of memory

        Examples:
            >>> con = ExaConnection(...)
            >>> myresult = con.export_to_pandas(
            ...    query_or_table="SELECT * FROM table"
            ... )
        """
        if not export_params:
            export_params = {}

        export_params["with_column_names"] = True

        return self.export_to_callback(
            cb.export_to_pandas,
            None,
            query_or_table,
            query_params,
            callback_params,
            export_params,
        )

    def export_to_parquet(
        self,
        dst: Union[Path, str],
        query_or_table: str,
        query_params: Optional[dict] = None,
        callback_params: Optional[dict] = None,
        export_params: Optional[dict] = None,
    ):
        """
        Export large amounts of data from Exasol to local parquet file(s).

        Args:
            dst:
                Local path to directory for exporting files. Can be one either a Path or
                str. **The default behavior, which can be changed via** ``callback_params``,
                **is that the specified directory should be empty.** If that is not
                the case, one of these exceptions may be thrown:

                    ValueError
                        '<dst>' exists and is not a directory
                    ValueError
                        '<dst>' contains existing files and `callback_params['existing_data_behavior']` is not one of these values: ("overwrite_or_ignore", "delete_matching").
                    pyarrow.lib.ArrowInvalid:
                        Could not write to <dst> Parquet Export from Exasol via Python Container/parquet as the directory is not empty and existing_data_behavior is to error
                    ValueError:
                        I/O operation on closed file.
                    DB error message:
                        ETL-5106: Following error occured while writing data to external connection [https://172.0.0.1:8653/000.csv failed after 200009 bytes. [OpenSSL SSL_read: SSL_ERROR_SYSCALL, errno 0],[56],[Failure when receiving data from the peer]] (Session: XXXXX)

                The ValueError exceptions would come from a check we provide via :func:`pyexasol.callback.check_export_to_parquet_directory_setting`.
                The purpose of calling this check is to detect issues before executing code within the callback pattern, which uses three threads.
                If a user has a different issue than we anticipated, it's possible the one of the other three exceptions is tossed for this, as
                discussed on `Importing and Exporting Data <https://exasol.github.io/pyexasol/master/user_guide/exploring_features/import_and_export/index.html>`__.

            query_or_table:
                SQL query or table from which to export data.
            query_params:
                Values for SQL query placeholders.
            callback_params:
                Dictionary with additional parameters for callback function
                `pyarrow.dataset.write_dataset <https://arrow.apache.org/docs/python/generated/pyarrow.dataset.write_dataset.html>`__.
                Some important defaults to note are:

                existing_data_behavior
                   Set to ``error``, which requires that the specified ``dst`` not
                   contain any files or an exception will be thrown.
                max_rows_per_file
                   Set to ``0``, which means that all rows will be written to 1 file.
                   If ``max_rows_per_file`` is altered, ensure that ``max_rows_per_group``
                   is set to a value less than or equal to the value of ``max_rows_per_file``.
                use_threads
                   Set to ``True`` and ``preserve_order`` is set to ``False``. This means
                   that the writing of multiple files will be done in parallel and that
                   the order is not guaranteed to be preserved.
            export_params:
                Custom parameters for EXPORT query.
        """
        if not export_params:
            export_params = {}

        cb.check_export_to_parquet_directory_setting(
            dst=dst, callback_params=callback_params
        )

        export_params["with_column_names"] = True

        return self.export_to_callback(
            cb.export_to_parquet,
            dst,
            query_or_table,
            query_params,
            callback_params,
            export_params,
        )

    def export_to_polars(
        self,
        query_or_table: str,
        query_params: Optional[dict] = None,
        callback_params: Optional[dict] = None,
        export_params: Optional[dict] = None,
    ) -> "polars.DataFrame":
        """
        Export large amount of data from Exasol to :class:`polars.DataFrame`.

        Args:
            query_or_table:
                SQL query or table from which to export data.
            query_params:
                Values for SQL query placeholders.
            callback_params:
                Dictionary with additional parameters for callback function
                `polars.read_csv <https://docs.pola.rs/api/python/stable/reference/api/polars.read_csv.html>`__.
            export_params:
                Custom parameters for EXPORT query.

        Returns:
            instance of :class:`polars.DataFrame`

        Warnings:
            - This function may run out of memory

        Examples:
            >>> con = ExaConnection(...)
            >>> df = con.export_to_polars(
            ...    query_or_table="SELECT * FROM table"
            ... )
        """
        if not export_params:
            export_params = {}

        export_params["with_column_names"] = True

        return self.export_to_callback(
            cb.export_to_polars,
            None,
            query_or_table,
            query_params,
            callback_params,
            export_params,
        )

    def import_from_file(self, src, table: str, import_params: Optional[dict] = None):
        """
        Import a large amount of data from a file or file-like object.

        Args:
            src:
                Source file or file-like object.
            table:
                Destination table for IMPORT.
            import_params:
                Custom parameters for IMPORT query.

        Note:
            File must be opened in binary mode.
        """
        return self.import_from_callback(
            cb.import_from_file, src, table, None, import_params
        )

    def import_from_iterable(
        self, src: Iterable, table: str, import_params: Optional[dict] = None
    ):
        """
        Import a large amount of data from an ``iterable`` Python object.

        Args:
            src:
                Source object implementing ``__iter__``.
                Iterator must return tuples of values.
            table:
                Destination table for IMPORT.
            import_params:
                Custom parameters for IMPORT query.
        """
        return self.import_from_callback(
            cb.import_from_iterable, src, table, None, import_params
        )

    def import_from_pandas(
        self,
        src: "pandas.DataFrame",
        table: str,
        callback_params: Optional[dict] = None,
        import_params: Optional[dict] = None,
    ):
        """
        Import a large amount of data from :class:`pandas.DataFrame`.

        Args:
            src:
                Source :class:`pandas.DataFrame` instance.
            table:
                Destination table for IMPORT.
            callback_params:
                Dictionary with additional parameters for callback function
                `pandas.DataFrame.to_csv <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_csv.html>`__.
            import_params:
                Custom parameters for IMPORT query.
        """
        return self.import_from_callback(
            cb.import_from_pandas, src, table, callback_params, import_params
        )

    def import_from_polars(
        self,
        src: Union["polars.LazyFrame", "polars.DataFrame"],
        table: str,
        callback_params: Optional[dict] = None,
        import_params: Optional[dict] = None,
    ):
        """
        Import a large amount of data from :class:`polars.DataFrame` or :class:`polars.LazyFrame`.

        Args:
            src:
                Source :class:`polars.DataFrame` or :class:`polars.LazyFrame` instance.
            table:
                Destination table for IMPORT.
            callback_params:
                Dictionary with additional parameters for callback function
                `polars.DataFrame.write_csv <https://docs.pola.rs/api/python/stable/reference/api/polars.DataFrame.write_csv.html>`__.
            import_params:
                Custom parameters for IMPORT query.
        """
        return self.import_from_callback(
            cb.import_from_polars, src, table, callback_params, import_params
        )

    def import_from_parquet(
        self,
        source: Union[list[Path], Path, str],
        table: str,
        callback_params: Optional[dict] = None,
        import_params: Optional[dict] = None,
    ):
        """
        Import a large amount of data from :class:`pyarrow.parquet.Table`.

        Args:
            source: Local filepath specification(s) to process. Can be one of:
                - list[pathlib.Path]: list of specific files
                - pathlib.Path: can be either a file or directory. If it's a directory,
                all files matching this pattern *.parquet will be processed.
                - str: representing a filepath which already contains a glob pattern
                (e.g., "/local_dir/*.parquet")
            table:
                Destination table for IMPORT.
            callback_params:
                Dict with additional parameters for callback function
                `parquet.ParquetFile.iter_batches <https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetFile.html#pyarrow.parquet.ParquetFile.iter_batches>`__.
            import_params:
                Custom parameters for IMPORT query.
        """
        return self.import_from_callback(
            cb.import_from_parquet, source, table, callback_params, import_params
        )

    def export_to_callback(
        self,
        callback: Callable,
        dst,
        query_or_table: str,
        query_params: Optional[dict] = None,
        callback_params: Optional[dict] = None,
        export_params: Optional[dict] = None,
    ):
        """
        Export a large amount of data to a user-defined callback function

        Args:
            callback:
                Callback function.
            dst:
                (optional) Path to file or file-like object where data will be exported to.
            query_or_table:
                SQL query or table from which to export data.
            query_params:
                Values for SQL query placeholders.
            callback_params:
                Dictionary with additional parameters for callback function.
            export_params:
                Custom parameters for EXPORT query.

        Returns:
            result of callback function

        Warnings:
            - This function may run out of memory

        Examples:
            >>> cb = lambda args: print(args)
            >>> con = ExaConnection(...)
            >>> con.export_to_callback(
            ...    callback=cb,
            ...    query_or_table="SELECT * FROM table"
            ... )
        """
        if not callable(callback):
            raise ValueError("Callback argument is not callable")

        if callback_params is None:
            callback_params = {}

        if export_params is None:
            export_params = {}

        if query_params is not None:
            query_or_table = self.format.format(query_or_table, **query_params)

        compression = (
            False if ("format" in export_params) else self.options["compression"]
        )

        http_thread = ExaHttpThread(
            self.ws_ipaddr,  # type: ignore
            self.ws_port,  # type: ignore
            compression,
            self.options["encryption"],
        )
        sql_thread = ExaSQLExportThread(
            self, compression, query_or_table, export_params
        )

        try:
            http_thread.start()

            sql_thread.set_http_thread(http_thread)
            sql_thread.start()

            with http_thread.read_pipe as pipe:
                result = callback(pipe, dst, **callback_params)

            http_thread.join_with_exc()
            sql_thread.join_with_exc()

            return result

        except (Exception, KeyboardInterrupt) as e:
            http_thread.terminate()
            http_thread.join()

            sql_thread.join(1)

            # Prevent infinite lock if SQL query is still running
            if sql_thread.is_alive():
                self.abort_query()
                sql_thread.join()

            # Give SQL exception higher priority
            if sql_thread.exc:
                raise sql_thread.exc

            raise e

    def import_from_callback(
        self,
        callback: Callable,
        src,
        table: str,
        callback_params: Optional[dict] = None,
        import_params: Optional[dict] = None,
    ):
        """
        Import a large amount of data from a user-defined callback function.

        Args:
            callback:
                Callback function.
            src:
                Source for the callback function.
            table:
                Destination table for IMPORT.
            callback_params:
                Dictionary with additional parameters for callback function.
            import_params:
                Custom parameters for IMPORT query.

        Raises:
            ValueError: callback argument isn't callable.
        """
        if callback_params is None:
            callback_params = {}

        if import_params is None:
            import_params = {}

        compression = (
            False if ("format" in import_params) else self.options["compression"]
        )

        if not callable(callback):
            raise ValueError("Callback argument is not callable")

        http_thread = ExaHttpThread(
            self.ws_ipaddr,  # type: ignore
            self.ws_port,  # type: ignore
            compression,
            self.options["encryption"],
        )
        sql_thread = ExaSQLImportThread(self, compression, table, import_params)

        try:
            http_thread.start()

            sql_thread.set_http_thread(http_thread)
            sql_thread.start()

            with http_thread.write_pipe as pipe:
                result = callback(pipe, src, **callback_params)

            http_thread.join_with_exc()
            sql_thread.join_with_exc()

            return result

        except (Exception, KeyboardInterrupt) as e:
            http_thread.terminate()
            http_thread.join()

            sql_thread.join(1)

            # Prevent infinite lock if SQL query is still running
            if sql_thread.is_alive():
                self.abort_query()
                sql_thread.join()

            # Give SQL exception higher priority
            if sql_thread.exc:
                raise sql_thread.exc

            raise e

    def export_parallel(
        self, exa_address_list, query_or_table, query_params=None, export_params=None
    ):
        """
        This function is part of :ref:`http_transport_parallel` API.

        Args:
            exa_address_list:
                List of ``ipaddr:port`` strings obtained from HTTP transport ``.address``.
            query_or_table:
                SQL query or table for the export.
            query_params:
                Values for SQL query placeholders.
            export_params:
                Custom parameters for Export query.
        Note:
            - Init HTTP transport in child processes first using pyexasol.http_transport()
            - Get internal Exasol address from each child process using .address
            - Pass address strings to parent process, combine into single list and use it for export_parallel() call
        """
        if export_params is None:
            export_params = {}

        compression = (
            False if ("format" in export_params) else self.options["compression"]
        )

        if query_params is not None:
            query_or_table = self.format.format(query_or_table, **query_params)

        # There is no need to actually run a separate thread here, all work is performed in separate processes
        # We simply reuse thread class to keep logic in one place
        sql_thread = ExaSQLExportThread(
            self, compression, query_or_table, export_params
        )
        sql_thread.set_exa_address_list(exa_address_list)
        sql_thread.run_sql()

    def import_parallel(self, exa_address_list, table, import_params=None):
        """
        This function is part of :ref:`http_transport_parallel` API.

        Args:
            exa_address_list:
                List of ``ipaddr:port`` strings obtained from HTTP transport ``.address``.
            table:
                Table to import to.
            import_params:
                Custom parameters for import.

        Note:
            - Init HTTP transport in child processes first using pyexasol.http_transport()
            - Get internal Exasol address from each child process using .address
            - Pass address strings to parent process, combine into single list and use it for import_parallel() call

        """
        if import_params is None:
            import_params = {}

        compression = (
            False if ("format" in import_params) else self.options["compression"]
        )

        # There is no need to actually run a separate thread here, all work is performed in separate processes
        # We simply reuse thread class to keep logic in one place
        sql_thread = ExaSQLImportThread(self, compression, table, import_params)
        sql_thread.set_exa_address_list(exa_address_list)
        sql_thread.run_sql()

    def session_id(self) -> str:
        """
        Session id of current session.

        Returns:
            Unique `SESSION_ID` of the current session as string.
        """
        return str(self.login_info.get("sessionId", ""))

    def protocol_version(self) -> int:
        """
        Actual protocol version used by the established connection.

        Returns:
            ``0`` if connection was not established yet (e.g. due to exception handling), otherwise protocol version as int.

        Warnings:
            Actual Protocol version might be downgraded from requested protocol version if Exasol server does not support it

        Note:
            The actual protocol version may be lower than the requested protocol version
            defined by the ``protocol_version`` connection option. For further details,
            refer to :ref:`protocol_version`.

        """
        return int(self.login_info.get("protocolVersion", 0))

    @property
    def exasol_db_version(self) -> Optional[Version]:
        """
        Version of the Exasol database of the current session.

        The login information is returned by the second response of LOGIN command
        and calls this "releaseVersion".
        """
        if release_version := self.login_info.get("releaseVersion"):
            return Version(release_version)
        return None

    def last_statement(self) -> ExaStatement:
        """
        Last created statement object

        Returns:
            ExaStatement: last created statement.

        Note:
            It is mainly used for HTTP transport to access internal IMPORT / EXPORT query,
            measure execution time and number of rows

        Tip:
            It is useful while working with `export_*` and `import_*` functions normally
            returning result of callback function instead of statement object.
        """
        if self.last_stmt is None:
            raise ExaRuntimeError(self, "Last statement not found")

        return self.last_stmt

    def close(self, disconnect=True):
        """
        Closes connection to database.

        Args:
            disconnect:
                If ``true`` send optional "disconnect" command to free resources and close session on Exasol server side properly.

        Note:
            Please note that "disconnect" should always be False when .close() is being called from .req()-like functions
            to prevent an infinite loop if websocket exception happens during handling of "disconnect" command
        """
        if self._ws.connected:
            if disconnect:
                self.req({"command": "disconnect"})

            self.logger.debug("[WebSocket connection close]")
            self._ws.close()

        self.is_closed = True
        self.last_stmt = None

    def get_attr(self):
        ret = self.req(
            {
                "command": "getAttributes",
            }
        )

        self.attr = ret["attributes"]

    def set_attr(self, new_attr):
        self.req(
            {
                "command": "setAttributes",
                "attributes": new_attr,
            }
        )

        # At this moment setAttributes response is inconsistent, so attributes must be refreshed after every call
        self.get_attr()

    def get_nodes(self, pool_size=None):
        """
        List of currently active Exasol nodes which is normally used for :ref:`http_transport_parallel`.

        Args:
            pool_size:
                Return list of specific size.

        Returns:
            list of dictionaries describing active Exasol nodes

        Note:

            Format: ``{'ipaddr': <ip_address>, 'port': <port>, 'idx': <incremental index of returned node>}``

            - If pool_size is bigger than number of nodes, list will wrap around and nodes will repeat with different 'idx'
            - If pool_size is omitted, return every active node once
            - It is useful to balance workload for parallel IMPORT and EXPORT Exasol shuffles list for every connection
            - Exasol shuffles list for every connection.
        """
        ret = self.req(
            {
                "command": "getHosts",
                "hostIp": self.ws_ipaddr,
            }
        )

        if pool_size is None:
            pool_size = ret["responseData"]["numNodes"]

        # Key 'host' is deprecated and remains only for backwards compatibility, please use `ipaddr` instead
        return [
            {"host": ipaddr, "ipaddr": ipaddr, "port": self.ws_port, "idx": idx}
            for idx, ipaddr in enumerate(
                itertools.islice(
                    itertools.cycle(ret["responseData"]["nodes"]), pool_size
                ),
                start=1,
            )
        ]

    def req(self, req):
        """Send WebSocket request and wait for response"""
        self.ws_req_count += 1
        local_req_count = self.ws_req_count

        # Build request
        send_data = self.json_encode(req)
        self.logger.debug_json(f"WebSocket request #{local_req_count}", req)

        # Prevent and discourage attempts to use connection object from another thread simultaneously
        if not self._req_lock.acquire(blocking=False):
            self.logger.debug(f"[WebSocket request #{local_req_count} WAS NOT SENT]")
            raise ExaConcurrencyError(
                self,
                "Connection cannot be shared between multiple threads "
                "sending requests simultaneously",
            )

        # Send request, wait for response
        try:
            start_ts = time.time()

            self._ws_send(send_data)
            recv_data = self._ws_recv()

            self.ws_req_time = time.time() - start_ts
        except (websocket.WebSocketException, ConnectionError) as e:
            self.close(disconnect=False)
            raise ExaCommunicationError(self, str(e))
        finally:
            self._req_lock.release()

        if not recv_data:
            raise ExaCommunicationError(
                self, "Empty WebSocket response, connection was likely closed"
            )

        # Parse response
        ret = self.json_decode(recv_data)
        self.logger.debug_json(f"WebSocket response #{local_req_count}", ret)

        # Updated attributes may be returned from any request
        if "attributes" in ret:
            self.attr = {**self.attr, **ret["attributes"]}

        if ret["status"] == "ok":
            return ret

        if ret["status"] == "error":
            # Special treatment for "execute" command to prevent very long tracebacks in most common cases
            if req.get("command") in ["execute", "createPreparedStatement"]:
                if ret["exception"]["sqlCode"] == "R0001":
                    cls_err = ExaQueryTimeoutError
                elif ret["exception"]["sqlCode"] == "R0003":
                    cls_err = ExaQueryAbortError
                else:
                    cls_err = ExaQueryError

                raise cls_err(
                    self,
                    req["sqlText"],
                    ret["exception"]["sqlCode"],
                    ret["exception"]["text"],
                )
            elif req.get("username") is not None:
                raise ExaAuthError(
                    self, ret["exception"]["sqlCode"], ret["exception"]["text"]
                )
            else:
                raise ExaRequestError(
                    self, ret["exception"]["sqlCode"], ret["exception"]["text"]
                )

    def abort_query(self):
        """
        Abort running query

        Warnings:

            This function should be called from a separate thread and has no response
            Response should be checked in the main thread which started execution of query

            There are three possible outcomes of calling this function:

            #. Query is aborted normally, connection remains active
            #. Query was stuck in a state which cannot be aborted, so Exasol has to terminate connection
            #. Query might be finished successfully before abort call had a chance to take effect
        """
        req = {"command": "abortQuery"}

        send_data = self.json_encode(req)
        self.logger.debug_json("WebSocket abort request", req)

        try:
            self._ws_send(send_data)
        except (websocket.WebSocketException, ConnectionError) as e:
            self.close(disconnect=False)
            raise ExaCommunicationError(self, str(e))

    def _login(self):
        start_ts = time.time()

        if self.options["access_token"] or self.options["refresh_token"]:
            auth_params = self._login_token()
        else:
            auth_params = self._login_password()

        self.login_info = self.req(
            {
                **auth_params,
                "driverName": f"{constant.DRIVER_NAME} {__version__}",
                "clientName": (
                    self.options["client_name"]
                    if self.options["client_name"]
                    else constant.DRIVER_NAME
                ),
                "clientVersion": (
                    self.options["client_version"]
                    if self.options["client_version"]
                    else __version__
                ),
                "clientOs": platform.platform(),
                "clientOsUsername": (
                    self.options["client_os_username"]
                    if self.options["client_os_username"]
                    else getpass.getuser()
                ),
                "clientRuntime": f"Python {platform.python_version()}",
                "useCompression": self.options["compression"],
                "attributes": self._get_login_attributes(),
            }
        )["responseData"]

        self.login_time = time.time() - start_ts

        if self.options["compression"]:
            self._ws_send = lambda x: self._ws.send_binary(
                zlib.compress(x.encode() if isinstance(x, str) else x, 1)
            )
            self._ws_recv = lambda: zlib.decompress(self._ws.recv())

    def _login_password(self):
        ret = self.req(
            {
                "command": "login",
                "protocolVersion": self.options["protocol_version"],
            }
        )

        auth_params = {
            "username": self.options["user"],
            "password": self._encrypt_password(ret["responseData"]["publicKeyPem"]),
        }

        return auth_params

    def _login_token(self):
        self.req(
            {
                "command": "loginToken",
                "protocolVersion": self.options["protocol_version"],
            }
        )

        auth_params = {}

        if self.options["refresh_token"]:
            auth_params["refreshToken"] = self.options["refresh_token"]

        if self.options["access_token"]:
            auth_params["accessToken"] = self.options["access_token"]

        return auth_params

    def _encrypt_password(self, public_key_pem):
        pk = rsa.PublicKey.load_pkcs1(public_key_pem.encode())
        return base64.b64encode(
            rsa.encrypt(self.options["password"].encode(), pk)
        ).decode()

    def _init_ws(self):
        """
        Init websocket connection

        Note:
            - Connection redundancy is supported
            - Specific Exasol node is randomly selected for every connection attempt
        """
        dsn_items = self._process_dsn(self.options["dsn"])
        failed_attempts = 0
        for hostname, ipaddr, port, fingerprint in dsn_items:
            try:
                self._ws = self._create_websocket_connection(
                    hostname, ipaddr, port, fingerprint
                )
            except Exception as e:
                failed_attempts += 1
                if failed_attempts == len(dsn_items):
                    raise ExaConnectionFailedError(
                        self, "Could not connect to Exasol: " + str(e)
                    ) from e
            else:
                self._ws.settimeout(self.options["socket_timeout"])

                self.ws_ipaddr = ipaddr or hostname
                self.ws_port = port

                self._ws_send = self._ws.send
                self._ws_recv = self._ws.recv

                if fingerprint:
                    self._validate_fingerprint(fingerprint)

                return

    def _create_websocket_connection(
        self, hostname: str, ipaddr: str, port: int, fingerprint: Optional[str]
    ) -> websocket.WebSocket:
        ws_options = self._get_ws_options(fingerprint=fingerprint)
        # Use correct hostname matching IP address for each connection attempt
        if self.options["encryption"] and self.options["resolve_hostnames"]:
            ws_options["sslopt"]["server_hostname"] = hostname

        connection_string = self._get_websocket_connection_string(
            hostname, ipaddr, port
        )
        self.logger.debug(f"Connection attempt {connection_string}")
        try:
            return websocket.create_connection(connection_string, **ws_options)
        except Exception as e:
            self.logger.debug(f"Failed to connect [{connection_string}]: {e}")
            raise e

    def _get_websocket_connection_string(
        self, hostname: str, ipaddr: Optional[str], port: int
    ) -> str:
        host = hostname
        if self.options["resolve_hostnames"]:
            if ipaddr is None:
                raise ValueError("IP address was not resolved")
            host = ipaddr
        if self.options["encryption"]:
            return f"wss://{host}:{port}"
        else:
            return f"ws://{host}:{port}"

    def _get_ws_options(self, fingerprint: Optional[str]) -> dict:
        options = {
            "timeout": self.options["connection_timeout"],
            "skip_utf8_validation": True,
            "enable_multithread": True,
            # Extra lock is necessary to protect abort_query() calls
        }

        if self.options["encryption"]:
            # refer to the `Security <https://exasol.github.io/pyexasol/master/user_guide/configuration/security.html>`__ page.
            if self.options["websocket_sslopt"] is None:
                # If a fingerprint is provided, then we do not use the default
                # to require a certificate verification.
                if fingerprint is not None:
                    options["sslopt"] = {"cert_reqs": ssl.CERT_NONE}
                else:
                    # When not provided by the user, the default behavior is to require
                    # strict certificate verification.
                    warn(
                        cleandoc(
                            """
                            From PyExasol version ``1.0.0``, the default behavior of
                            ExaConnection for encrypted connections without a fingerprint
                            is to require strict certificate validation with
                            ``websocket_sslopt=None`` being mapped to
                            ``{"cert_reqs": ssl.CERT_REQUIRED}``. The prior default behavior
                            was to map such cases to ``{"cert_reqs": ssl.CERT_NONE}``. For
                            more information about encryption & best practices, please refer to
                            `Security <https://exasol.github.io/pyexasol/master/user_guide/configuration/security.html>`__ page.
                            """
                        ),
                        PyexasolWarning,
                    )
                    options["sslopt"] = {"cert_reqs": ssl.CERT_REQUIRED}
            else:
                options["sslopt"] = self.options["websocket_sslopt"].copy()

        if self.options["http_proxy"]:
            proxy_components = urllib.parse.urlparse(self.options["http_proxy"])

            if proxy_components.hostname is None:
                raise ValueError("Could not parse http_proxy")

            options["http_proxy_host"] = proxy_components.hostname
            options["http_proxy_port"] = proxy_components.port
            options["http_proxy_auth"] = (
                proxy_components.username,
                proxy_components.password,
            )

        return options

    def _get_login_attributes(self):
        attributes = {
            "currentSchema": str(self.options["schema"]),
            "autocommit": self.options["autocommit"],
            "queryTimeout": self.options["query_timeout"],
        }

        if self.options["snapshot_transactions"] is not None:
            attributes["snapshotTransactionsEnabled"] = self.options[
                "snapshot_transactions"
            ]

        return attributes

    def _process_dsn(self, dsn: str) -> list[Host]:
        """
        Parse DSN, expand ranges and resolve IP addresses for all hostnames.

        Note:
            Randomness is required to guarantee proper distribution of workload across all nodes

        Returns:
            List of (hostname, ip_address, port) tuples in random order
        """
        if dsn is None or len(dsn.strip()) == 0:
            raise ExaConnectionDsnError(self, "Connection string is empty")

        current_port = constant.DEFAULT_PORT
        current_fingerprint = None

        result = []

        dsn_re = re.compile(
            r"^(?P<hostname_prefix>.+?)"
            # Optional range (e.g. myxasol1..4.com)
            r"(?:(?P<range_start>\d+)\.\.(?P<range_end>\d+)(?P<hostname_suffix>.*?))?"
            # Optional fingerprint (e.g. myexasol1..4.com/135a1d2dce102de866f58267521f4232153545a075dc85f8f7596f57e588a181)
            r"(?:/(?P<fingerprint>[0-9A-Fa-f]+|nocertcheck))?"
            # Optional port (e.g. myexasol1..4.com:8564)
            r"(?::(?P<port>\d+)?)?$",
            re.IGNORECASE,
        )

        # Port is applied backwards, so we iterate the whole list backwards to avoid second loop
        for part in reversed(dsn.split(",")):
            if len(part) == 0:
                continue

            m = dsn_re.search(part)

            if not m:
                raise ExaConnectionDsnError(
                    self, f"Could not parse connection string part [{part}]"
                )

            # Optional port
            if m.group("port"):
                current_port = int(m.group("port"))

            # Optional fingerprint
            if m.group("fingerprint"):
                current_fingerprint = m.group("fingerprint").upper()

                if not self.options["encryption"]:
                    raise ExaConnectionDsnError(
                        self,
                        "Fingerprint was specified in connection string, but encryption is not enabled",
                    )

            # Hostname or IP range was specified, expand it
            if m.group("range_start"):
                if int(m.group("range_start")) > int(m.group("range_end")):
                    raise ExaConnectionDsnError(
                        self,
                        f"Connection string part [{part}] contains an invalid range, "
                        f"lower bound is higher than upper bound",
                    )

                zfill_width = len(m.group("range_start"))

                for i in range(
                    int(m.group("range_start")), int(m.group("range_end")) + 1
                ):
                    hostname = f"{m.group('hostname_prefix')}{str(i).zfill(zfill_width)}{m.group('hostname_suffix')}"
                    result.extend(
                        self._resolve_hostname(
                            hostname, current_port, current_fingerprint
                        )
                    )
            # Just a single hostname or single IP address
            else:
                hostname = m.group("hostname_prefix")
                if self.options["resolve_hostnames"]:
                    result.extend(
                        self._resolve_hostname(
                            hostname, current_port, current_fingerprint
                        )
                    )
                else:
                    result.append(
                        Host(hostname, None, current_port, current_fingerprint)
                    )

        random.shuffle(result)

        return result

    def _resolve_hostname(
        self, hostname: str, port: int, fingerprint: Optional[str]
    ) -> list[Host]:
        """
        Resolve all IP addresses for hostname and add port.

        Warnings:
            - It also implicitly checks that all hostnames mentioned in DSN can be resolved
        """
        try:
            hostname, _, ipaddr_list = socket.gethostbyname_ex(hostname)
        except OSError as e:
            raise ExaConnectionDsnError(
                self,
                f"Could not resolve IP address of hostname [{hostname}] "
                f"derived from connection string",
            ) from e

        return [Host(hostname, ipaddr, port, fingerprint) for ipaddr in ipaddr_list]

    def _validate_fingerprint(self, provided_fingerprint):
        if provided_fingerprint.upper() != "NOCERTCHECK":
            server_fingerprint = (
                hashlib.sha256(self._ws.sock.getpeercert(True)).hexdigest().upper()
            )

            if provided_fingerprint != server_fingerprint:
                raise ExaConnectionFailedError(
                    self,
                    f"Provided fingerprint [{provided_fingerprint}] did not match "
                    f"server fingerprint [{server_fingerprint}]",
                )

    def _init_logger(self):
        self.logger = self.cls_logger(self, constant.DRIVER_NAME)
        self.logger.setLevel("DEBUG" if self.options["debug"] else "WARNING")
        self.logger.add_default_handler()

    def _init_format(self):
        self.format = self.cls_formatter(self)

    def _init_json(self):
        if self.options["json_lib"] == "rapidjson":
            import rapidjson

            self.json_encode = lambda x, indent=False: rapidjson.dumps(
                x,
                number_mode=rapidjson.NM_NATIVE,
                indent=2 if indent else None,
                ensure_ascii=False,
            )
            self.json_decode = lambda x: rapidjson.loads(
                x, number_mode=rapidjson.NM_NATIVE
            )

        elif self.options["json_lib"] == "ujson":
            import ujson

            self.json_encode = lambda x, indent=False: ujson.dumps(
                x, indent=2 if indent else 0, ensure_ascii=False
            )
            self.json_decode = lambda x: ujson.loads(x)

        elif self.options["json_lib"] == "orjson":
            import orjson

            self.json_encode = lambda x, indent=False: orjson.dumps(
                x,
                option=orjson.OPT_INDENT_2 | orjson.OPT_APPEND_NEWLINE if indent else 0,
            )
            self.json_decode = lambda x: orjson.loads(x)

        elif self.options["json_lib"] == "json":
            import json

            self.json_encode = lambda x, indent=False: json.dumps(
                x, indent=2 if indent else None, ensure_ascii=False
            )
            self.json_decode = lambda x: json.loads(x)

        else:
            raise ValueError(f"Unsupported json library [{self.options['json_lib']}]")

    def _init_ext(self):
        self.ext = self.cls_extension(self)

    def _init_meta(self):
        self.meta = self.cls_meta(self)

    def _get_stmt_output_dir(self):
        import pathlib
        import tempfile

        if self.options["udf_output_dir"]:
            base_output_dir = self.options["udf_output_dir"]
        else:
            base_output_dir = tempfile.gettempdir()

        # Create unique sub-directory for every statement of every session
        self._udf_output_count += 1

        stmt_output_dir = (
            pathlib.Path(base_output_dir)
            / f"{self.session_id()}_{self._udf_output_count}"
        )
        stmt_output_dir.mkdir(parents=True, exist_ok=True)

        return stmt_output_dir

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} session_id={self.session_id()}"
            f" dsn={self.options['dsn']} user={self.options['user']}>"
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        """
        Will close the connection.

        Note:
            close() is being called automatically in order to:

            #.  send OP_CLOSE frame to Exasol server rather than silently terminating the socket on client side
            #.  make sure connection is closed immediately even if garbage collection was disabled for any reasons
            #.  write debug logs
        """
        # Based on our investigations, two scenarios have emerged, one of which does not function correctly:
        #
        # 1. `__del__` is invoked during a regular garbage collector run while the process remains active.
        # 2. `__del__` is invoked during interpreter shutdown (throws an exception).
        #     * This leads to a minor residual on the database, which will automatically be cleared up within a couple of hours
        #
        # In situations where interpreter shutdown occurs while a connection remains unclosed
        # due to not being addressed during a regular garbage collector run or manual close,
        # the error scenario (2.) will be triggered.
        #
        # The issue arises because during interpreter shutdown, orderly deletion is not guaranteed
        # (see also [Python documentation](https://docs.python.org/3/reference/datamodel.html#object.__del__) and
        # [C-API documentation](https://docs.python.org/3/c-api/init.html#c.Py_FinalizeEx)).
        # Consequently, scenarios may occur where the underlying socket is already closed or cleaned up while
        # the connection still tries using it (see [#108](https://github.com/exasol/pyexasol/issues/108)).
        #
        # Despite this problem, having a `__del__` method that performs cleanup remains valuable,
        # particularly in the context of long-running processes, to prevent resource leaks.
        #
        # Possible alternative solutions to address this would be:
        # * Within the catch, report a clear error message that the user did not close all the connections which were opened.
        # * Remove the `__del__` function to "force" the user to always address open/close in a clean manner.
        #
        # Still, the feedback we received from users so far indicates that neither alternative is desirable for them.
        try:
            self.close()
        except Exception:
            pass



"""
This module provides `PEP-249`_ compliant DBAPI exceptions.
(see also `PEP-249-exceptions`_)

.. _PEP-249-exceptions: https://peps.python.org/pep-0249/#exceptions
"""


class Warning(Exception):  # Required by spec. pylint: disable=W0622
    """
    Exception raised for important warnings like data truncations while inserting, etc.
    """


class Error(Exception):
    """
    Base class of all other error exceptions.
    You can use this to catch all errors with one single except statement.
    Warnings are not considered errors and thus should not use this class as base.
    """


class InterfaceError(Error):
    """
    Exception raised for errors that are related to the database interface rather than
    the database itself.
    """


class DatabaseError(Error):
    """Exception raised for errors that are related to the database."""


class DataError(DatabaseError):
    """
    Exception raised for errors that are due to problems with the processed
    data like division by zero, numeric value out of range, etc.
    """


class OperationalError(DatabaseError):
    """
    Exception raised for errors that are related to the database’s operation
    and not necessarily under the control of the programmer, e.g. an unexpected
    disconnect occurs, the data source name is not found, a transaction
    could not be processed, a memory allocation error occurred during processing, etc.
    """


class IntegrityError(DatabaseError):
    """
    Exception raised when the relational integrity of the database is affected,
    e.g. a foreign key check fails.
    """


class InternalError(DatabaseError):
    """
    Exception raised when the database encounters an internal error,
    e.g. the cursor is not valid anymore, the transaction is out of sync, etc.
    """


class ProgrammingError(DatabaseError):
    """
    Exception raised for programming errors, e.g. table not found or already exists,
    syntax error in the SQL statement, wrong number of parameters specified, etc.
    """


class NotSupportedError(DatabaseError):
    """
    Exception raised in case a method or database API was used which is not supported
    by the database, e.g. requesting a .rollback() on a connection that does not
    support transaction or has transactions turned off.
    """



import configparser
import pathlib


class ExaLocalConfig:
    """
    Parse local config file
    Prepare arguments for :func:`pyexasol.connect` method

    Arguments accepting functions class names and tuples are not supported
    You may pass such arguments directly in code using `**kwargs`
    """

    default_filename = ".pyexasol.ini"

    arg_types = {
        "autocommit": bool,
        "snapshot_transactions": bool,
        "socket_timeout": int,
        "query_timeout": int,
        "compression": bool,
        "encryption": bool,
        "fetch_dict": bool,
        "fetch_size_bytes": int,
        "lower_ident": bool,
        "quote_ident": bool,
        "verbose_error": bool,
        "debug": bool,
        "udf_output_port": int,
    }

    def __init__(self, config_path=None):
        if config_path:
            self.path = pathlib.Path(config_path)
        else:
            self.path = self.get_default_path()

        if not self.path.exists():
            raise RuntimeError(f"PyExasol config file [{self.path}] not found")

        self.parser = configparser.ConfigParser()
        self.parser.read(self.path, encoding="utf-8")

    def get_args(self, section):
        args = {}

        if not self.parser.has_section(section):
            raise ValueError(
                f"Section [{section}] in PyExasol config file [{self.path}] not found"
            )

        for k in self.parser[section]:
            t = self.get_arg_type(k)

            if t == bool:
                args[k] = self.parser[section].getboolean(k)
            elif t == int:
                args[k] = self.parser[section].getint(k)
            elif t == float:
                args[k] = self.parser[section].getfloat(k)
            else:
                args[k] = self.parser[section].get(k)

        return args

    def get_default_path(self):
        return pathlib.Path.home() / self.default_filename

    def get_arg_type(self, k):
        return self.arg_types.get(k, str)



class PyexasolWarning(UserWarning):
    """Base class for all warnings emitted by pyexasol."""


class PyexasolDeprecationWarning(PyexasolWarning, DeprecationWarning):
    """Warn about features that will be removed in future versions."""



import re
import string


class ExaFormatter(string.Formatter):
    """
        :class:`pyexasol.ExaFormatter` is a subclass of :class:`string.Formatter` designed to prevent SQL injections in Exasol dynamic SQL queries.

        Note:
            It introduces set of placeholders to prevent SQL injections specifically
            in Exasol dynamic SQL queries. It also completely disables `format_spec`
            section of standard formatting since it has no use in context of
            SQL queries and may cause more harm than good.

    You may access these functions using `.format` property of connection object. Example:

        Examples:

            >>> C = pyexasol.connect(...)
            ... print(C.format.escape('abc'))
    """

    safe_ident_regexp = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")
    safe_decimal_regexp = re.compile(r"^(\+|-)?[0-9]+(\.[0-9]+)?$")
    safe_float_regexp = re.compile(r"^(\+|-)?[0-9]+(\.[0-9]+((e|E)(\+|-)[0-9]+)?)?$")

    def __init__(self, connection):
        self.connection = connection

        self.conversions = {
            "s": self.quote,
            "d": self.safe_decimal,
            "f": self.safe_float,
            "i": self.safe_ident,
            "q": self.quote_ident,
            "r": str,
        }

        self.default_conversion = "s"

        # Set default treatment for identifiers passed as strings to relevant functions
        if self.connection.options["quote_ident"]:
            self.default_format_ident = (
                self.quote_ident
            )  # Identifiers will be quoted and escaped
            self.default_format_ident_value = (
                str  # Identifier values will be left unchanged
            )
        else:
            self.default_format_ident = (
                self.safe_ident
            )  # Identifiers will only be checked for safety
            self.default_format_ident_value = (
                str.upper
            )  # Identifier values will be transformed to upper-case

    def format_field(self, value, format_spec):
        if format_spec != "":
            raise ValueError("format_spec is disabled for ExaFormatter")

        return value

    def convert_field(self, value, conversion):
        if conversion is None:
            conversion = self.default_conversion

        if conversion not in self.conversions:
            raise ValueError(f"Unknown conversion {conversion}")

        if isinstance(value, list):
            if not value:
                raise ValueError("Trying to format an empty list")
            return ", ".join([self.conversions[conversion](v) for v in value])
        else:
            return self.conversions[conversion](value)

    @classmethod
    def escape(cls, val):
        """
        Takes a raw value and converts it into an and escaped string.

        Args:
            val: Value to be escaped.

        Returns:
            A string where all single quotes ``'`` have been replaced
            with two single quotes ``''``.
        """
        return str(val).replace("'", "''")

    @classmethod
    def escape_ident(cls, val):
        """
        Takes a raw value and converts it into an and escaped string.

        Args:
            val: Value to be escaped.

        Returns:
            A string where all double quotes ``"`` have been replaced
            with two double quotes ``""``.
        """
        return str(val).replace('"', '""')

    @classmethod
    def escape_like(cls, val):
        """
        Escape LIKE-patterns.

        Args:
            val: Value to be escaped.

        Returns:
            A string where all double quotes ``\\`` have been replaced
            with ``\\\\``, where ``%`` have been replaced with ``\\%``,
            where ``_`` have been replaced with ``\\_``.
        """
        return (
            cls.escape(val)
            .replace("\\", "\\\\")
            .replace("%", r"\%")
            .replace("_", r"\_")
        )

    @classmethod
    def quote(cls, val):
        """
        Escapes a string using :meth:`pyexasol.ExaFormatter.escape` and wraps it in single quotes ``'``.
        """
        if val is None:
            return "NULL"

        return f"'{cls.escape(val)}'"

    @classmethod
    def quote_ident(cls, val):
        """
        Escapes an object or a tuple of objects using :meth:`pyexasol.ExaFormatter.escape_ident` and wraps it in double quotes ``"``.

        Args:
            val: Raw identifier(s) to be escaped.

        Returns:
            str: The formatted and quoted identifier, or joined identifiers if
            a tuple was provided.
        """
        if isinstance(val, tuple):
            return ".".join([cls.quote_ident(x) for x in val])

        return f'"{cls.escape_ident(val)}"'

    @classmethod
    def safe_ident(cls, val):
        """
        Convert a raw identifier safely.

        Args:
            val (str or tuple): Raw identifier(s).

        Returns:
            Validates identifier as string.

        Raises:
            ValueError If passed values is not a valid identifier (e.g. contains spaces)

        Warning:
            It puts it into SQL query without any quotting.
        """
        if isinstance(val, tuple):
            return ".".join([cls.safe_ident(x) for x in val])

        val = str(val)

        if not cls.safe_ident_regexp.match(val):
            if "." in val:
                parts = val.split(".")
                raise ValueError(
                    f"Value [{val}] is not a safe identifier. Please use tuple to pass schema names. "
                    f"Example: ('{parts[0]}', '{parts[1]}')"
                )
            elif '"' in val:
                raise ValueError(
                    f"Value [{val}] is not a safe identifier. Please use 'quote_ident' or '!q' conversion "
                    f"to pass identifiers with lowercase or special characters."
                )
            else:
                raise ValueError(f"Value [{val}] is not a safe identifier")

        return val

    @classmethod
    def safe_float(cls, val):
        """
        Convert a float safely to string.

        Args:
            val: Float value to convert.

        Returns:
            Validates identifier as string.

        Raises:
            ValueError: If value is not valid, e.g.: ``+infinity`` or ``-infinity``.
        """
        if val is None:
            return "NULL"

        val = str(val)

        if not cls.safe_float_regexp.match(val):
            raise ValueError(f"Value [{val}] is not a safe float")

        return val

    @classmethod
    def safe_decimal(cls, val):
        """
        Convert a decimal safely to string.

        Args:
            val: Decimal value to convert.

        Returns:
            Validates identifier as string.

        Raises:
            ValueError: If value is not valid.
        """
        if val is None:
            return "NULL"

        val = str(val)

        if not cls.safe_decimal_regexp.match(val):
            raise ValueError(f"Value [{val}] is not a safe integer")

        return val

    def __repr__(self):
        return f"<{self.__class__.__name__} session_id={self.connection.session_id()}>"



import collections
import itertools

from . import constant
from .exceptions import ExaRuntimeError


class ExaStatement:
    """
    This class executes and helps to fetch result set of single Exasol SQL statement.

    Warning:
        Unlike typical `Cursor` object, `ExaStatement` is not reusable.

    Note:
        :class:`pyexasol.ExaStatement` may fetch result set rows as ``tuples`` (default)
        or as ``dict`` (set `fetch_dict=True` in connection options).

        :class:`pyexasol.ExaStatement` may use custom data-type mapper during fetching
        (set `fetch_mapper=<func>` in connection options).
        Mapper function accepts two arguments (raw `value` and `dataType` object)
        and returns custom object or value.

        :class:`pyexasol.ExaStatement` fetches big result sets in chunks.
        The size of chunk may be adjusted (set `fetch_size_bytes=<int>` in connection options).

        Public Attributes:
            ``execution_time``:
                Execution time of SQL statement. It is measured by wall-clock time
                of WebSocket request, so real execution time is a bit faster.
    """

    def __init__(
        self,
        connection,
        query=None,
        query_params=None,
        prepare=False,
        meta_nosql=False,
        **options,
    ):
        """
        Args:
            connection:
                -
            query:
                -
            query_params:
                -
            prepare:
                -
            meta_nosql:
                -
            options:
                additonal kwargs
        """
        self.connection = connection

        self.query = query if meta_nosql else self._format_query(query, query_params)
        self.query_params = query_params

        self.fetch_dict = options.get(
            "fetch_dict", self.connection.options["fetch_dict"]
        )
        self.fetch_mapper = options.get(
            "fetch_mapper", self.connection.options["fetch_mapper"]
        )
        self.fetch_size_bytes = options.get(
            "fetch_size_bytes", self.connection.options["fetch_size_bytes"]
        )
        self.lower_ident = options.get(
            "lower_ident", self.connection.options["lower_ident"]
        )

        self.data_zip = zip()
        self.col_names = []
        self.col_types = []

        self.num_columns = 0
        self.num_rows_total = 0
        self.num_rows_chunk = 0

        self.row_count = 0
        self.pos_total = 0
        self.pos_chunk = 0

        self.result_type = None
        self.result_set_handle = None
        self.statement_handle = None
        self.parameter_data = None

        if self.connection.is_closed:
            raise ExaRuntimeError(self.connection, "Exasol connection was closed")

        # Always set last_stmt in connection object regardless of how statement object was created
        self.connection.last_stmt = self

        # This index may not match STMT_ID in system tables due to automatically executed queries (e.g. autocommit)
        self.connection.stmt_count += 1
        self.stmt_idx = self.connection.stmt_count

        self.execution_time = 0
        self.is_closed = False

        if prepare:
            self._prepare()
        elif meta_nosql:
            self._execute_meta_nosql()
        else:
            self._execute()

    def __iter__(self):
        """
        The best way to fetch result set of statement is to use iterator:

        Yields:
            ``tuple`` or ``dict`` depending on ``fetch_dict`` connection option.

        Examples:

            >>> st = pyexasol.execute('SELECT * FROM table')
            ... for row in st:
            ...     print(row)
        """
        return self

    def __next__(self):
        if self.pos_total >= self.num_rows_total:
            if self.result_type != "resultSet":
                raise ExaRuntimeError(
                    self.connection,
                    "Attempt to fetch from statement without result set",
                )

            self._close_result_set_handle()
            raise StopIteration

        if self.pos_chunk >= self.num_rows_chunk:
            self._next_chunk()

        row = next(self.data_zip)

        if self.fetch_mapper:
            row = tuple(map(self.fetch_mapper, row, self.col_types))

        if self.fetch_dict:
            row = dict(zip(self.col_names, row))

        self.pos_total += 1
        self.pos_chunk += 1

        return row

    def fetchone(self):
        """
        Fetches one row of data.

        Returns:
            ``tuple`` or ``dict``.
            ``None`` if all rows were fetched.
        """
        try:
            row = next(self)
        except StopIteration:
            return None

        return row

    def fetchmany(self, size=constant.DEFAULT_FETCHMANY_SIZE):
        """
        Fetch multiple rows.

        Args:
            size:
                Set the specific number of rows to fetch (Default: ``10000``)

        Returns:
            ``list`` of ``tuples`` or ``list`` of ``dict``.
            Empty `list` if all rows were fetched previously.
        """
        return [row for row in itertools.islice(self, size)]

    def fetchall(self):
        """
        Fetches all remaining rows.

        Returns:
            ``list`` of ``tuples`` or ``list`` of ``dict``.
            Empty ``list`` if all rows were fetched previously.

        Warning:
            This function may exhaust available memory.
        """
        return [row for row in self]

    def fetchcol(self):
        """
        Fetches all values from the first column.

        Returns:
            ``list`` of values.
            Empty ``list`` if all rows were fetched previously.
        """
        self.fetch_dict = False
        return [row[0] for row in self]

    def fetchval(self):
        """
        Fetches first column of first row.


        Returns:
            Value, ``None`` if all rows were fetched previously.

        Tip:
            This may be useful for queries returning single value like
            ``SELECT count(*) FROM table``.
        """
        self.fetch_dict = False

        try:
            row = next(self)
        except StopIteration:
            return None

        return row[0]

    def rowcount(self):
        """
        Number of selected/processed rows.

        Returns:
            Total amount of selected rows for statements with result set (``num_rows``).
            Total amount of processed rows for DML queries (``row_count``).
        """
        if self.result_type == "resultSet":
            return self.num_rows_total
        else:
            return self.row_count

    def columns(self):
        """
        Retrieves column information of returned data.

        Returns:
            A ``dict`` with keys as ``column names`` and values as ``dataType`` objects.

        Notes:

            The dict will containt the following data:

            .. list-table::
               :header-rows: 1

               * - Names
                 - Type
                 - Description
               * - type
                 - string
                 - column data type
               * - precision
                 - number
                 - (optional) column precision
               * - scale
                 - number
                 - (optional) column scale
               * - size
                 - number
                 - (optional) maximum size in bytes of a column value
               * - characterSet
                 - string
                 - (optional) character encoding of a text column
               * - withLocalTimeZone
                 - true, false
                 - (optional) specifies if a timestamp has a local time zone
               * - fraction
                 - number
                 - (optional) fractional part of number
               * - srid
                 - number
                 - (optional) spatial reference system identifier
        """
        return dict(zip(self.col_names, self.col_types))

    def column_names(self):
        """List of column names."""
        return self.col_names

    def close(self):
        """
        Closes result set handle if it was opened.

        Warning:
            You won't be able to fetch next chunk of large dataset
            after calling this function, but no other side effects.
        """
        self._close_result_set_handle()
        self._close_statement_handle()

        self.is_closed = True

    def _close_result_set_handle(self):
        if not self.connection.is_closed and self.result_set_handle:
            self.connection.req(
                {
                    "command": "closeResultSet",
                    "resultSetHandles": [self.result_set_handle],
                }
            )

            self.result_set_handle = None

    def _close_statement_handle(self):
        if not self.connection.is_closed and self.statement_handle:
            self.connection.req(
                {
                    "command": "closePreparedStatement",
                    "statementHandle": self.statement_handle,
                }
            )

            self.statement_handle = None

    def _format_query(self, query, query_params):
        query = str(query)

        if query_params is not None:
            query = self.connection.format.format(query, **query_params)

        return query.lstrip(" \n").rstrip(" \n;")

    def _execute(self):
        ret = self.connection.req(
            {
                "command": "execute",
                "sqlText": self.query,
            }
        )

        self.execution_time = self.connection.ws_req_time
        self._init_result_set(ret)

    def _execute_meta_nosql(self):
        meta_params = self.query_params if self.query_params is not None else {}

        if "command" in meta_params:
            raise ExaRuntimeError(
                self.connection,
                "Key 'command' is not allowed as a parameter for meta nosql request",
            )

        if "attributes" in meta_params:
            raise ExaRuntimeError(
                self.connection,
                "Key 'attributes' is not allowed as a parameter for meta nosql request",
            )

        ret = self.connection.req(
            {
                "command": self.query,
                **meta_params,
            }
        )

        self.execution_time = self.connection.ws_req_time
        self._init_result_set(ret)

    def _prepare(self):
        ret = self.connection.req(
            {
                "command": "createPreparedStatement",
                "sqlText": self.query,
            }
        )

        self.statement_handle = ret["responseData"]["statementHandle"]

        if "parameterData" in ret["responseData"]:
            self.parameter_data = ret["responseData"]["parameterData"]

        self._init_result_set(ret)

    def execute_prepared(self, data=None):
        ret = self.connection.req(
            {
                "command": "executePreparedStatement",
                "statementHandle": self.statement_handle,
                "numColumns": (
                    self.parameter_data["numColumns"] if self.parameter_data else 0
                ),
                "numRows": len(data) if data else 0,
                "columns": (
                    self.parameter_data["columns"] if self.parameter_data else []
                ),
                "data": list(zip(*data)) if data else [],
            }
        )

        self.execution_time = self.connection.ws_req_time
        self._init_result_set(ret)

    def _init_result_set(self, ret):
        res = ret["responseData"]["results"][0]

        self.result_type = res["resultType"]

        if self.result_type == "resultSet":
            if "resultSetHandle" in res["resultSet"]:
                self.result_set_handle = res["resultSet"]["resultSetHandle"]

            if "data" in res["resultSet"]:
                self.data_zip = zip(*res["resultSet"]["data"])

            if self.lower_ident:
                self.col_names = [
                    c["name"].lower() for c in res["resultSet"]["columns"]
                ]
            else:
                self.col_names = [c["name"] for c in res["resultSet"]["columns"]]

            self.col_types = [c["dataType"] for c in res["resultSet"]["columns"]]

            self.num_columns = res["resultSet"]["numColumns"]
            self.num_rows_total = res["resultSet"]["numRows"]
            self.num_rows_chunk = res["resultSet"]["numRowsInMessage"]

            self._check_duplicate_col_names()
        elif self.result_type == "rowCount":
            self.row_count = res["rowCount"]
        else:
            raise ExaRuntimeError(
                self.connection, f"Unknown resultType: {self.result_type}"
            )

    def _next_chunk(self):
        ret = self.connection.req(
            {
                "command": "fetch",
                "resultSetHandle": self.result_set_handle,
                "startPosition": self.pos_total,
                "numBytes": self.fetch_size_bytes,
            }
        )

        if "data" in ret["responseData"]:
            self.data_zip = zip(*ret["responseData"]["data"])
        else:
            self.data_zip = zip()

        self.num_rows_chunk = ret["responseData"]["numRows"]
        self.pos_chunk = 0

    def _check_duplicate_col_names(self):
        """
        Exasol allows duplicate names in result sets, but it leads to various problems related to dictionaries
        PyExasol adds additional check to prevent such problems and to allow safe .columns() and fetch_dict=True
        """
        duplicate_col_names = [
            k for (k, v) in collections.Counter(self.col_names).items() if v > 1
        ]

        if duplicate_col_names:
            raise ExaRuntimeError(
                self.connection,
                f'Duplicate column names in result set: {", ".join(duplicate_col_names)}',
            )

    def __repr__(self):
        return f"<{self.__class__.__name__} session_id={self.connection.session_id()} stmt_idx={self.stmt_idx}>"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        """
        close() is being called automatically in order to close handles which
        might be still opened on Exasol server side
        """
        try:
            self.close()
        except Exception:
            pass



"""
This module provides `PEP-249`_ DBAPI compliant cursor implementation.
(see also `PEP-249-cursor`_)

.. _PEP-249-cursor: https://peps.python.org/pep-0249/#cursor-objects
"""

import datetime
import decimal
from collections import defaultdict
from dataclasses import (
    astuple,
    dataclass,
)
from functools import wraps
from typing import Optional

import pyexasol.exceptions
from exasol.driver.websocket._errors import (
    Error,
    NotSupportedError,
)
from exasol.driver.websocket._types import TypeCode


@dataclass
class MetaData:
    """Metadata describing a result column"""

    name: str
    type_code: TypeCode
    display_size: Optional[int] = None
    internal_size: Optional[int] = None
    precision: Optional[int] = None
    scale: Optional[int] = None
    null_ok: Optional[bool] = None


def _pyexasol2dbapi_metadata(name, metadata) -> MetaData:
    type_mapping = {t.value: t for t in TypeCode}
    key_mapping = {
        "name": "name",
        "type_code": "type",
        "precision": "precision",
        "scale": "scale",
        "display_size": "unknown",
        "internal_size": "size",
        "null_ok": "unknown",
    }
    metadata = defaultdict(lambda: None, metadata)
    metadata["type"] = type_mapping[metadata["type"]]
    metadata["name"] = name
    return MetaData(**{k: metadata[v] for k, v in key_mapping.items()})


def _is_not_closed(method):
    """
    Mark a function to require an open connection.

    Raises:
        An Error if the marked function is called without an open connection.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self._is_closed:
            raise Error(
                f"Unable to execute operation <{method.__name__}>, because cursor was already closed."
            )
        return method(self, *args, **kwargs)

    return wrapper


def _requires_result(method):
    """
    Decorator requires the object to have a result.

    Raises:
        Error if the cursor object has not produced a result yet.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self._cursor:
            raise Error("No result has been produced.")
        return method(self, *args, **kwargs)

    return wrapper


def _identity(value):
    return value


def _pyexasol2dbapi(value, metadata):
    members = (
        "name",
        "type_code",
        "display_size",
        "internal_size",
        "precision",
        "scale",
        "null_ok",
    )
    metadata = MetaData(**{k: v for k, v in zip(members, metadata)})

    def to_date(v):
        if not isinstance(v, str):
            return v
        return datetime.date.fromisoformat(v)

    def to_float(v):
        if not isinstance(v, str):
            return v
        return float(v)

    converters = defaultdict(
        lambda: _identity,
        {
            TypeCode.Date: to_date,
            TypeCode.Double: to_float,
        },
    )
    converter = converters[metadata.type_code]
    return converter(value)


def _dbapi2pyexasol(value):
    converters = defaultdict(
        lambda: _identity,
        {decimal.Decimal: str, float: str, datetime.date: str, datetime.datetime: str},
    )
    converter = converters[type(value)]
    return converter(value)


class Cursor:
    """
    Implementation of a cursor based on the DefaultConnection.

    For more details see :class: `Cursor` protocol definition.
    """

    # see https://peps.python.org/pep-0249/#arraysize
    DBAPI_DEFAULT_ARRAY_SIZE = 1

    def __init__(self, connection):
        self._connection = connection
        self._cursor = None
        self._is_closed = False

    # Note: Needed for compatibility with sqlalchemy_exasol base dialect.
    def __enter__(self):
        return self

    # Note: Needed for compatibility with sqlalchemy_exasol base dialect.
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    @_is_not_closed
    def arraysize(self):
        """See also :py:meth: `Cursor.arraysize`"""
        return self.DBAPI_DEFAULT_ARRAY_SIZE

    @property
    @_is_not_closed
    def description(self):
        """See also :py:meth: `Cursor.description`"""
        if not self._cursor:
            return None
        columns_metadata = (
            _pyexasol2dbapi_metadata(name, metadata)
            for name, metadata in self._cursor.columns().items()
        )
        columns_metadata = tuple(astuple(metadata) for metadata in columns_metadata)
        columns_metadata = columns_metadata if columns_metadata != () else None
        return columns_metadata

    @property
    @_is_not_closed
    def rowcount(self):
        """
        See also :py:meth: `Cursor.rowcount`

        Attention: This implementation of the rowcount attribute deviates slightly
                   from what the dbapi2 requires.

               Difference:

                    If the rowcount of the last operation cannot be determined it will
                    return 0.

                    Expected by DBAPI2 = -1
                    Actually returned  = 0

               Rational:

                    With the usage of pyexasol as underlying driver, there is no trivial
                    way to do this.
        """
        if not self._cursor:
            return -1
        return self._cursor.rowcount()

    @_is_not_closed
    def callproc(self, procname, parameters=None):
        """See also :py:meth: `Cursor.callproc`"""
        raise NotSupportedError("Optional and therefore not supported")

    @_is_not_closed
    def close(self):
        """See also :py:meth: `Cursor.close`"""
        self._is_closed = True
        if not self._cursor:
            return
        self._cursor.close()

    @_is_not_closed
    def execute(self, operation, parameters=None):
        """See also :py:meth: `Cursor.execute`"""
        if parameters:
            self.executemany(operation, [parameters])
            return

        connection = self._connection.connection
        try:
            self._cursor = connection.execute(operation)
        except pyexasol.exceptions.ExaError as ex:
            raise Error() from ex

    @staticmethod
    def _adapt_to_requested_db_types(parameters, db_response):
        """
        Adapt parameter types to match the types requested by the DB in the
        `createPreparedStatement <https://github.com/exasol/websocket-api/blob/master/docs/commands/createPreparedStatementV1.md>`_
        response.

        Args:

            parameters: which will be passed/send to the database.
            db_response: contains the DB response including the required types.


        Attention:

            This shim method currently only patches the following types:

                * VARCHAR
                * DOUBLE

            therefore it the future it may be necessary to improve or extend this.

            A hint that patching of a specific type is required, could be and
            error message similar to this one:

            .. code-block::

                pyexasol.exceptions.ExaRequestError:
                ...
                message  => getString: JSON value is not a string. (...)
                ...
        """

        def varchar(value):
            if value is None:
                return None
            return str(value)

        def double(value):
            if value is None:
                return None
            return float(value)

        converters = defaultdict(
            lambda: _identity, {"VARCHAR": varchar, "DOUBLE": double}
        )
        selected_converters = (
            converters[column["dataType"]["type"]] for column in db_response["columns"]
        )
        parameters = zip(selected_converters, parameters)
        parameters = [converter(value) for converter, value in parameters]
        return parameters

    @_is_not_closed
    def executemany(self, operation, seq_of_parameters):
        """See also :py:meth: `Cursor.executemany`"""
        parameters = [
            [_dbapi2pyexasol(p) for p in params] for params in seq_of_parameters
        ]
        connection = self._connection.connection
        self._cursor = connection.cls_statement(connection, operation, prepare=True)

        parameter_data = self._cursor.parameter_data
        parameters = [
            Cursor._adapt_to_requested_db_types(params, parameter_data)
            for params in parameters
        ]
        try:
            self._cursor.execute_prepared(parameters)
        except pyexasol.exceptions.ExaError as ex:
            raise Error() from ex

    def _convert(self, rows):
        if rows is None:
            return None

        return tuple(self._convert_row(row) for row in rows)

    def _convert_row(self, row):
        if row is None:
            return row

        return tuple(
            _pyexasol2dbapi(value, metadata)
            for value, metadata in zip(row, self.description)
        )

    @_requires_result
    @_is_not_closed
    def fetchone(self):
        """See also :py:meth: `Cursor.fetchone`"""
        row = self._cursor.fetchone()
        return self._convert_row(row)

    @_requires_result
    @_is_not_closed
    def fetchmany(self, size=None):
        """See also :py:meth: `Cursor.fetchmany`"""
        size = size if size is not None else self.arraysize
        rows = self._cursor.fetchmany(size)
        return self._convert(rows)

    @_requires_result
    @_is_not_closed
    def fetchall(self):
        """See also :py:meth: `Cursor.fetchall`"""
        rows = self._cursor.fetchall()
        return self._convert(rows)

    @_is_not_closed
    def nextset(self):
        """See also :py:meth: `Cursor.nextset`"""
        raise NotSupportedError("Optional and therefore not supported")

    @_is_not_closed
    def setinputsizes(self, sizes):
        """See also :py:meth: `Cursor.setinputsizes`

        Attention:
            This method does nothing.
        """

    @_is_not_closed
    def setoutputsize(self, size, column):
        """See also :py:meth: `Cursor.setoutputsize`

        Attention:
            This method does nothing.
        """

    def __del__(self):
        if self._is_closed:
            return
        self.close()



"""
Extension with Exasol-specific helper functions
"""

from typing import Optional

from .exceptions import ExaRuntimeError


class ExaExtension:
    """
    This class extends the functionality of a simple SQL driver to address common Exasol-related problems.

    Tip:
        You may access these functions using `.ext` property of connection object.

    Examples:
        >>> C = pyexasol.connect(...)
        ... print(C.ext.get_disk_space_usage())
    """

    def __init__(self, connection):
        self.connection = connection
        self.reserved_words = None

    def get_columns(self, object_name):
        """
        Get information about columns of table or view.

        Args:
            object_name: Object name may be passed as tuple to specify custom schema.

        Caution:
            **DEPRECATED**, please use ``.meta.sql_columns`` instead.
        """
        object_name = self.connection.format.default_format_ident(object_name)
        return self.get_columns_sql(f"SELECT * FROM {object_name}")

    def get_columns_sql(self, query, query_params=None):
        """
        Get columns of SQL query without executing it.

        Args:
            object_name: Object name may be passed as tuple to specify custom schema.

        Caution:
            **DEPRECATED**, please use ``.meta.sql_columns`` instead.

        Note:
            It relies on a prepared statement that will be closed immediately without execution
        """
        stmt = self.connection.cls_statement(
            self.connection, query, query_params, prepare=True
        )
        columns = stmt.columns()
        stmt.close()

        return columns

    def insert_multi(self, table_name, data, columns=None):
        """
        Insert a small number of rows into a table using a prepared statement.

        Args:
            table_name:
                Target table for INSERT.
            data:
                Source object implementing ``__iter__`` (e.g. list or tuple).
            columns:
                List of column names to specify custom order of columns.

        Tip:
            Compared to ``.import_from_iterable``, this method offers better performance for small data sets of 10,000 rows or fewer.

            * Use ``.import_from_iterable`` for larger data sets and better memory efficiency
            * Use ``.import_from_pandas`` to import from :class:`pandas.DataFrame` regardless of its size
            * Use ``.import_from_parquet`` to import from parquet files
            * Use ``.import_from_polars`` to import from :class:`polars.DataFrame` regardless of its size

            You may use "columns" argument to specify custom order of columns for insertion
            If some columns are not included in this list, ``NULL`` or ``DEFAULT`` value will be used instead

        Note:
            Please note that data should be presented in a row format. You may use ``zip(*data_cols)`` to convert columnar format into row format.
        """

        # Convert possible iterator into list
        data = list(data)

        if len(data) == 0:
            raise ExaRuntimeError(
                self.connection,
                "At least one row of data is required for insert_multi()",
            )

        params = {
            "table_name": self.connection.format.default_format_ident(table_name),
            "columns": "",
            "values": ", ".join(["?"] * len(data[0])),
        }

        if columns:
            params["columns"] = (
                f"({','.join([self.connection.format.default_format_ident(c) for c in columns])})"
            )

        query = "INSERT INTO {table_name!r}{columns!r} VALUES ({values!r})"

        stmt = self.connection.cls_statement(
            self.connection, query, params, prepare=True
        )
        stmt.execute_prepared(data)
        stmt.close()

        return stmt

    def get_sys_columns(self, object_name):
        """
        Get information about columns of table or view (SYS format)

        Args:
            object_name: Object name may be passed as tuple to specify custom schema.

        Caution:
            **DEPRECATED**, please use ``.meta.list_columns`` instead.
        """
        if isinstance(object_name, tuple):
            schema = self.connection.format.default_format_ident_value(object_name[0])
            object_name = self.connection.format.default_format_ident_value(
                object_name[1]
            )
        else:
            schema = self.connection.current_schema()
            object_name = self.connection.format.default_format_ident_value(object_name)

        sql = """/*snapshot execution*/
            SELECT c.column_name, c.column_type, c.column_maxsize, c.column_num_scale,
                   c.column_is_nullable, c.column_is_distribution_key, c.column_default,
                   c.column_comment, t.type_name
            FROM EXA_ALL_COLUMNS c
                JOIN EXA_SQL_TYPES t ON (c.column_type_id=t.type_id)
            WHERE c.column_schema={schema}
                AND c.column_table={object_name}
            ORDER BY c.column_ordinal_position
        """

        st = self._execute(sql, {"schema": schema, "object_name": object_name})
        res = list()

        for r in st:
            res.append(
                {
                    "name": r["column_name"],
                    "type": r["type_name"],
                    "sql_type": r["column_type"],
                    "size": r["column_maxsize"],
                    "scale": r["column_num_scale"],
                    "nulls": r["column_is_nullable"],
                    "distribution_key": r["column_is_distribution_key"],
                    "default": r["column_default"],
                    "comment": r["column_comment"],
                }
            )

        return res

    def get_sys_tables(self, schema=None, table_name_prefix=""):
        """
        Get information about tables in selected schema(SYS format)

        Args:
            schema:
                -
            table_name_prefix:
                Output may be optionally filtered by table name prefix.

        Caution:
            **DEPRECATED**, please use ``.meta.list_tables`` instead.
        """
        if schema is None:
            schema = self.connection.current_schema()
        else:
            schema = self.connection.format.default_format_ident_value(schema)

        table_name_prefix = self.connection.format.default_format_ident_value(
            table_name_prefix
        )
        table_name_prefix = self.connection.format.escape_like(table_name_prefix)

        sql = """/*snapshot execution*/
            SELECT *
            FROM EXA_ALL_TABLES
            WHERE table_schema={schema}
                AND table_name LIKE '{table_name_prefix!r}%'
            ORDER BY table_name ASC
        """

        st = self._execute(
            sql, {"schema": schema, "table_name_prefix": table_name_prefix}
        )
        res = list()

        for r in st:
            res.append(
                {
                    "table_name": (
                        r["table_name"].lower()
                        if self.connection.options["lower_ident"]
                        else r["table_name"]
                    ),
                    "table_schema": (
                        r["table_schema"].lower()
                        if self.connection.options["lower_ident"]
                        else r["table_schema"]
                    ),
                    "table_is_virtual": r["table_is_virtual"],
                    "table_has_distribution_key": r["table_has_distribution_key"],
                    "table_comment": r["table_comment"],
                }
            )

        return res

    def get_sys_views(self, schema: Optional[str] = None, view_name_prefix: str = ""):
        """
        Get information about views in selected schema (SYS format)

        Args:
            schema:
                -
            view_name_prefix:
                Output may be optionally filtered by view name prefix.

        Caution:
            **DEPRECATED**, please use ``.meta.list_views`` instead.
        """
        if schema is None:
            schema = self.connection.current_schema()
        else:
            schema = self.connection.format.default_format_ident_value(schema)

        view_name_prefix = self.connection.format.default_format_ident_value(
            view_name_prefix
        )
        view_name_prefix = self.connection.format.escape_like(view_name_prefix)

        sql = """/*snapshot execution*/
            SELECT *
            FROM EXA_ALL_VIEWS
            WHERE view_schema={schema}
                AND view_name LIKE '{view_name_prefix!r}%'
            ORDER BY view_name ASC
        """

        st = self._execute(
            sql, {"schema": schema, "view_name_prefix": view_name_prefix}
        )
        res = list()

        for r in st:
            res.append(
                {
                    "view_name": (
                        r["view_name"].lower()
                        if self.connection.options["lower_ident"]
                        else r["view_name"]
                    ),
                    "view_schema": (
                        r["view_schema"].lower()
                        if self.connection.options["lower_ident"]
                        else r["view_schema"]
                    ),
                    "scope_schema": (
                        r["scope_schema"].lower()
                        if self.connection.options["lower_ident"]
                        else r["scope_schema"]
                    ),
                    "view_text": r["view_text"],
                    "view_comment": r["view_comment"],
                }
            )

        return res

    def get_sys_schemas(self, schema_name_prefix: str = ""):
        """
        Get information about schemas (SYS format)

        Args:
            schema_name_prefix:
                Output may be optionally filtered by schema name prefix

        Caution:
            **DEPRECATED**, please use ``.meta.list_schemas`` instead.
        """
        schema_name_prefix = self.connection.format.default_format_ident_value(
            schema_name_prefix
        )
        schema_name_prefix = self.connection.format.escape_like(schema_name_prefix)

        sql = """/*snapshot execution*/
            SELECT *
            FROM EXA_SCHEMAS
            WHERE schema_name LIKE '{schema_name_prefix!r}%'
            ORDER BY schema_name ASC
        """

        st = self._execute(sql, {"schema_name_prefix": schema_name_prefix})
        res = list()

        for r in st:
            res.append(
                {
                    "schema_name": (
                        r["schema_name"].lower()
                        if self.connection.options["lower_ident"]
                        else r["schema_name"]
                    ),
                    "schema_owner": (
                        r["schema_owner"].lower()
                        if self.connection.options["lower_ident"]
                        else r["schema_owner"]
                    ),
                    "schema_is_virtual": r["schema_is_virtual"],
                    "schema_comment": r["schema_comment"],
                }
            )

        return res

    def get_reserved_words(self):
        """
        Get reserved keywords which cannot be used as identifiers without double-quotes.

        Caution:
            **DEPRECATED**, please use ``.meta.list_sql_keywords`` instead.

        Warning:
            Never hard-code this list! It changes with every Exasol versions.
        """
        if self.reserved_words is None:
            sql = """
                SELECT keyword
                FROM EXA_SQL_KEYWORDS
                WHERE reserved IS TRUE
                ORDER BY keyword
            """

            self.reserved_words = self._execute(sql).fetchcol()

        return self.reserved_words

    def get_disk_space_usage(self):
        """
        Get the disk space usage of the Exasol DB.

        Returns:

            A dict with 4 keys, providing all disk space details.

            .. list-table::
               :header-rows: 1

               * - Key
                 - Description
               * - ``occupied_size``
                 - How much space is occupied (in bytes)
               * - ``free_size``
                 - How much space is available (in bytes)
               * - ``total_size``
                 - occupied_size + free_size
               * - ``occupied_size_percent``
                 - Percentage of occupied disk space (0-100%)

        Note:
            Exasol still lacks a standard function to measure actual disk space usage.
            We are trying to mitigate this problem by creating a custom function.
        """
        sql = """
            SELECT measure_time,
                   (committed_size * redundancy + temp_swap_data) AS occupied_size,
                   (device_size * redundancy + hdd_free) AS total_size
            FROM "$EXA_STATS_DB_SIZE"
            ORDER BY measure_time DESC
            LIMIT 1
        """

        row = self._execute(sql).fetchone()

        if row is None:
            return None

        row["occupied_size"] = int(row["occupied_size"])
        row["total_size"] = int(row["total_size"])

        row["free_size"] = row["total_size"] - row["occupied_size"]
        row["occupied_size_percent"] = round(
            row["occupied_size"] / row["total_size"] * 100, 2
        )

        return row

    def export_to_pandas_with_dtype(self, query_or_table, query_params=None):
        """
        Export to pandas and attempt to guess correct dtypes based on Exasol columns.

        Args:
            query_or_table:
                Query or table to export.
            query_params:
                Additional query parameters.

        Note:
            Since pandas has significantly smaller range of allowed values, this function makes many assumptions.
            Please use it as baseline for your own function for complex cases.

            .. list-table::
               :widths: 25 25
               :header-rows: 1

               * - Exasol Type
                 - Pandas Type
               * - Small decimal
                 - int32
               * - Big decimal
                 - int64
               * - Double
                 - float64
               * - Date, Timestamp
                 - datetime64[ns]
               * - Everything else
                 - category (!)
        """

        if query_params:
            query_or_table = self.connection.format.format(
                query_or_table, **query_params
            )

        if (
            isinstance(query_or_table, tuple)
            or str(query_or_table).strip().find(" ") == -1
        ):
            columns = self.get_columns(query_or_table)
        else:
            columns = self.get_columns_sql(query_or_table)

        params = {
            "names": list(),
            "dtype": dict(),
            "parse_dates": list(),
            "na_values": dict(),
            "infer_datetime_format": True,
            "engine": "c",
            "skip_blank_lines": False,
        }

        for k, c in columns.items():
            params["names"].append(k)

            if c["type"] == "DATE":
                params["dtype"][k] = "object"
                params["na_values"][k] = "0001-01-01"
                params["parse_dates"].append(k)

            elif c["type"] == "TIMESTAMP":
                params["dtype"][k] = "object"
                params["na_values"][k] = "0001-01-01 00:00:00"
                params["parse_dates"].append(k)

            elif c["type"] == "DECIMAL":
                if c["scale"] > 0:
                    params["dtype"][k] = "float64"
                else:
                    if c["precision"] <= 9:
                        params["dtype"][k] = "int32"
                    else:
                        params["dtype"][k] = "int64"

            elif c["type"] == "DOUBLE":
                params["dtype"][k] = "float64"

            else:
                params["dtype"][k] = "category"

        def callback(pipe, dst, **kwargs):
            import pandas

            return pandas.read_csv(pipe, **kwargs)

        return self.connection.export_to_callback(
            callback, None, query_or_table, None, params
        )

    def explain_last(self, details: bool = False):
        """
        Args:
            details (bool):
                - ``False``, the function returns the average (AVG) or maximum (MAX) values aggregated for all Exasol nodes.
                - ``True``, the function returns separate rows for each individual Exasol node, with a column labeled "iproc" representing the node.

        Returns:
            Profiling information for last executed query.

        Note:
            This function should be called immediately after ``execute()``
            ``COMMIT``, ``ROLLBACK`` and ``FLUSH STATISTICS`` queries are ignored.
        Tip:
            Details are useful to detect bad data distribution and imbalanced execution
            If you want to see real values of ``CPU, MEM, HDD, NET`` columns,
            please enable Exasol profiling first with:

            .. code-block:: sql

                ALTER SESSION SET PROFILE = 'ON';

        *Please refer to Exasol User Manuals for explanations about profiling columns.*
        """
        self._execute("FLUSH STATISTICS")

        sql = """
            SELECT part_id /* PyExasol explain_last */
                {iproc_col!r}
                , part_name
                , part_info
                , object_schema
                , object_name
                , object_rows
                , in_rows
                , out_rows
                , duration
                , start_time
                , stop_time
                , cpu
                , mem_peak
                , temp_db_ram_peak
                , hdd_read
                , hdd_write
                , net
                , remarks
            FROM {table_name!q}
            WHERE session_id=CURRENT_SESSION
                AND stmt_id = CURRENT_STATEMENT - {stmt_offset!d}
            ORDER BY {order_by!r}
        """

        params = {
            "iproc_col": ", iproc" if details else "",
            "table_name": (
                "$EXA_PROFILE_DETAILS_LAST_DAY" if details else "$EXA_PROFILE_LAST_DAY"
            ),
            "order_by": "part_id ASC, iproc ASC" if details else "part_id ASC",
            "stmt_offset": 4 if self.connection.attr["autocommit"] else 2,
        }

        return self._execute(sql, params).fetchall()

    def _execute(self, query, query_params=None):
        # Preserve ext-functions output format regardless of current options for user queries
        options = {
            "fetch_dict": True,
            "fetch_mapper": None,
            "lower_ident": True,
        }

        return self.connection.cls_statement(
            self.connection, query, query_params, **options
        )

    def __repr__(self):
        return f"<{self.__class__.__name__} session_id={self.connection.session_id()}>"



"""
Collection of basic callbacks for import / export functions
You may write your own callbacks for complex cases

--------------

Export callback arguments:
1) pipe opened in "rb" mode
2) destination for callback (if applicable, can be None)
3) callback_params as dict

--------------

Import callback arguments:
1) pipe opened in "wb" mode
2) source data for callback (e.g. pandas data frame, list, file object)
3) callback_params as dict

"""

import csv
import glob
import io
import shutil
from collections.abc import Iterable
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Optional,
    Union,
)

if TYPE_CHECKING:
    import pandas
    import polars


def export_to_list(pipe, dst, **kwargs) -> list:
    """
    Basic example of how to export CSV stream into basic list of tuples
    """
    wrapped_pipe = io.TextIOWrapper(pipe, newline="\n", encoding="utf-8")
    reader = csv.reader(wrapped_pipe, lineterminator="\n", **kwargs)

    return list(reader)


def export_to_pandas(pipe, dst, **kwargs) -> "pandas.DataFrame":
    """
    Basic example of how to export into :class:`pandas.DataFrame`
    Custom params for :func:`pandas.DataFrame.read_csv` may be passed in `**kwargs`
    """
    import pandas

    return pandas.read_csv(pipe, skip_blank_lines=False, **kwargs)


def check_export_to_parquet_directory_setting(
    dst: Union[Path, str], callback_params: Optional[dict] = None
) -> None:
    """
    Check that the dst directory is in an allowed state:
      - does not yet exist
      - does exist but is empty
      - does exist, is not empty, and the callback_params['existing_data_behavior'] is "overwrite_or_ignore" or "delete_matching"
    If this is not the case, then an exception is raised.

    Without this function, when the aforementioned conditions were not met and
    :meth:`pyexasol.ExaConnection.export_to_parquet` was executed, then an exception would be
    raised:
      - `FileExistsError: [Errno 17] Cannot create directory '<dst>': non-directory entry exists. Detail: [errno 17] File exists`
      - `pyarrow.lib.ArrowInvalid: Could not write to <dst> Parquet Export from Exasol via
    Python Container/parquet as the directory is not empty and existing_data_behavior
    is to error`. However, due to the three threads used in the underlying callback
    pattern, it is possible that another final exception would be raised, as
    discussed on `Importing and Exporting Data <https://exasol.github.io/pyexasol/master/user_guide/exploring_features/import_and_export/index.html>`__.
    """
    dir_path = Path(dst)
    if not dir_path.exists():
        return

    if not dir_path.is_dir():
        raise ValueError(f"'{dst}' exists and is not a directory")

    if not callback_params:
        callback_params = {}

    existing_data_behavior = callback_params.get("existing_data_behavior")
    allowed_existing_data_behavior = ("overwrite_or_ignore", "delete_matching")
    if existing_data_behavior not in allowed_existing_data_behavior:
        if any(dir_path.iterdir()):
            raise ValueError(
                f"'{dst}' contains existing files and `callback_params['existing_data_behavior']` is not one of these values: {allowed_existing_data_behavior}."
            )


def export_to_parquet(pipe, dst: Union[Path, str], **kwargs) -> None:
    """
    Basic example of how to export into local parquet file(s)

    Args:
        dst:
            Local path to directory for exporting files. Can be one either a Path or
            str. **The default behavior, which can be changed via** `**kwargs`,
            **is that the specified directory should be empty.** If that is not the case,
            one of these exceptions may be thrown:

                    pyarrow.lib.ArrowInvalid:
                        Could not write to <dst> Parquet Export from Exasol via Python Container/parquet as the directory is not empty and existing_data_behavior is to error
                    ValueError:
                        I/O operation on closed file.
                    DB error message:
                        ETL-5106: Following error occured while writing data to external connection [https://172.0.0.1:8653/000.csv failed after 200009 bytes. [OpenSSL SSL_read: SSL_ERROR_SYSCALL, errno 0],[56],[Failure when receiving data from the peer]] (Session: XXXXX)

            To better detect this earlier on, it is recommended to use
            :func:`pyexasol.callback.check_export_to_parquet_directory_setting`.

        **kwargs:
            Custom params for :func:`pyarrow.dataset.write_dataset`. Some important
            defaults to note are:

              existing_data_behavior
                  Set to ``error``, which requires that the specified ``dst`` not
                  contain any files or an exception will be thrown.
              max_rows_per_file
                  Set to ``0``, which means that all rows will be written to 1 file.
                  If ``max_rows_per_file`` is altered, ensure that ``max_rows_per_group``
                  is set to a value less than or equal to the value of ``max_rows_per_file``.
              use_threads
                  Set to ``True`` and ``preserve_order`` is set to ``False``. This means
                  that the writing of multiple files will be done in parallel and that
                  the order is not guaranteed to be preserved.
    """
    from pyarrow import (
        csv,
        dataset,
    )

    parse_options = csv.ParseOptions(newlines_in_values=True)
    read_options = csv.ReadOptions(use_threads=False)
    reader = csv.open_csv(pipe, read_options=read_options, parse_options=parse_options)
    dataset.write_dataset(
        reader,
        base_dir=dst,
        format="parquet",
        use_threads=False,
        preserve_order=True,
        **kwargs,
    )


def export_to_polars(pipe, dst, **kwargs) -> "polars.DataFrame":
    """
    Basic example of how to export into :class:`polars.DataFrame`
    Custom params for :func:`polars.read_csv` may be passed in `**kwargs`
    """
    import polars

    return polars.read_csv(pipe, **kwargs)


def export_to_file(pipe, dst):
    """
    Basic example of how to export into file or file-like object opened in binary mode
    """
    if not hasattr(dst, "write"):
        dst = open(dst, "wb")

    shutil.copyfileobj(pipe, dst, 65535)


def import_from_iterable(pipe, src: Iterable, **kwargs):
    """
    Basic example of how to import from iterable object (list, dict, tuple, iterator, generator, etc.)
    """
    if not hasattr(src, "__iter__"):
        raise ValueError("Data source is not iterable")

    wrapped_pipe = io.TextIOWrapper(pipe, newline="\n", encoding="utf-8")
    writer = csv.writer(
        wrapped_pipe, lineterminator="\n", quoting=csv.QUOTE_NONNUMERIC, **kwargs
    )

    for row in src:
        writer.writerow(row)


def import_from_pandas(pipe, src: "pandas.DataFrame", **kwargs):
    """
    Basic example of how to import from :class:`pandas.DataFrame`
    Custom params for :fun:`pandas.DataFrame.to_csv` may be passed in **kwargs
    """
    import pandas

    if not isinstance(src, pandas.DataFrame):
        raise ValueError("Data source is not pandas.DataFrame")

    wrapped_pipe = io.TextIOWrapper(pipe, newline="\n", encoding="utf-8")
    src.to_csv(
        wrapped_pipe,
        header=False,
        index=False,
        lineterminator="\n",
        quoting=csv.QUOTE_NONNUMERIC,
        **kwargs,
    )


def get_parquet_files(source: Union[list[Path], Path, str]) -> list[Path]:
    if isinstance(source, str):
        matches = glob.glob(source)
        return sorted(
            [filepath for i in matches if (filepath := Path(i)) and filepath.is_file()]
        )
    if isinstance(source, Path):
        if source.is_file():
            return [source]
        elif source.is_dir():
            return sorted(source.glob("*.parquet"))
    if isinstance(source, list):
        not_a_valid_path_file = [
            filepath
            for filepath in source
            if not isinstance(filepath, Path) or not filepath.is_file()
        ]
        if len(not_a_valid_path_file) > 0:
            raise ValueError(
                f"source {source} contained entries which were not `Path` or valid `Path` files {not_a_valid_path_file}"
            )
        return source
    raise ValueError(
        f"source {source} is not a supported type (Union[list[Path], Path, str])."
    )


def import_from_parquet(
    pipe, source: Union[list[Path], Path, str], **kwargs
):  # NOSONAR(S3776)
    """
    Basic example of how to import from :class:`pyarrow.parquet.ParquetFile` via local parquet file(s)

    Args:
        source: Local filepath specification(s) to process. Can be one of:
            - list[pathlib.Path]: list of specific files
            - pathlib.Path: can be either a file or directory. If it is a directory,
            all files matching the following pattern *.parquet will be processed.
            - str: representing a filepath which already contains a glob pattern
            (e.g., "/local_dir/*.parquet")
        **kwargs:
            Custom params for :func:`pyarrow.parquet.ParquetFile.iter_batches`. This can be used
            to specify what columns should be read and their preferred order.

    Please note that nested or hierarchical column types are not supported.
    """
    from pyarrow import (
        csv,
        parquet,
        types,
    )

    def ensure_no_nested_columns(
        schema, requested_columns: Union[list[str], None]
    ) -> None:
        nested_fields = []
        for field in schema:
            if not types.is_nested(field.type):
                continue
            if requested_columns and field.name in requested_columns:
                nested_fields.append(field)
            if not requested_columns:
                nested_fields.append(field)

        if nested_fields:
            raise ValueError(
                f"Fields {nested_fields} of schema from file {file} is hierarchical which is not supported."
            )

    if not (parquet_files := get_parquet_files(source)):
        raise ValueError(f"source {source} does not match any files")

    columns = kwargs.get("columns", None)
    for file in parquet_files:
        parquet_file = parquet.ParquetFile(file, memory_map=True)
        ensure_no_nested_columns(parquet_file.schema_arrow, columns)
        for batch in parquet_file.iter_batches(**kwargs):
            write_options = csv.WriteOptions(include_header=False)
            csv.write_csv(batch, pipe, write_options=write_options)


def import_from_polars(
    pipe, src: Union["polars.LazyFrame", "polars.DataFrame"], **kwargs
):
    """
    Basic example of how to import from :class:`polars.DataFrame` or :class:`polars.LazyFrame`
    Custom params for :func:`polars.DataFrame.write_csv` may be passed in `**kwargs`
    """
    import polars

    if isinstance(src, polars.LazyFrame):
        src = src.collect()
    elif not isinstance(src, polars.DataFrame):
        raise ValueError("Data source is not polars.DataFrame or polars.LazyFrame")

    return src.write_csv(pipe, include_header=False, **kwargs)


def import_from_file(pipe, src):
    """
    Basic example of how to import from file or file-like object opened in binary mode
    """
    if not hasattr(src, "read"):
        src = open(src, "rb")

    shutil.copyfileobj(src, pipe, 65536)



import logging
import subprocess
import uuid


class DockerDataLoader:
    """Data loader for docker based Exasol DB"""

    def __init__(self, dsn, username, password, container_name, data_directory):
        self._logger = logging.getLogger("DockerDataLoader")
        self._dsn = dsn
        self._user = username
        self._password = password
        self._container = container_name
        self._data_directory = data_directory
        self._tmp_dir = f"data-{uuid.uuid4()}"

    @property
    def data_directory(self):
        return self._data_directory

    @property
    def ddl_file(self):
        return self.data_directory / "import.sql"

    @property
    def csv_files(self):
        return self.data_directory.rglob("*.csv")

    def load(self):
        self._create_dir()
        self._upload_files()
        self._import_data()

    def _execute_command(self, command):
        self._logger.info("Executing docker command: %s", command)
        result = subprocess.run(
            command,
            check=True,
            text=True,
            capture_output=True,
        )
        self._logger.debug("Stderr: %s", result.stderr)
        return result.stdout

    def _exaplus(self) -> str:
        # Note that the directory for the binary might change between DB versions
        find_exaplus = [
            "docker",
            "exec",
            self._container,
            "find",
            "/opt",
            "/usr",
            "-name",
            "exaplus",
            "-type",
            "f",  # only files
            "-executable",  # only executable files
            "-print",  # -print -quit will stop after the result is found
            "-quit",
        ]
        exaplus = self._execute_command(find_exaplus).strip()
        self._logger.info("Found exaplus at %s", exaplus)
        return exaplus

    def _create_dir(self):
        """Create data directory within the docker container."""
        mkdir = ["docker", "exec", self._container, "mkdir", self._tmp_dir]
        stdout = self._execute_command(mkdir)
        self._logger.info("Stdout: %s", stdout)

    def _upload_files(self):
        files = [self.ddl_file]
        files.extend(self.csv_files)
        for file in files:
            copy_file = [
                "docker",
                "cp",
                f"{file.resolve()}",
                f"{self._container}:{self._tmp_dir}/{file.name}",
            ]
            stdout = self._execute_command(copy_file)
            self._logger.debug("Stdout: %s", stdout)

    def _import_data(self):
        """Load test data into a backend."""
        execute_ddl_file = [
            "docker",
            "exec",
            "-w",
            f"/{self._tmp_dir}",
            self._container,
            self._exaplus(),
            "-c",
            f"{self._dsn}",
            "-u",
            self._user,
            "-p",
            self._password,
            "-f",
            self.ddl_file.name,
            "--jdbcparam",
            "validateservercertificate=0",
        ]
        stdout = self._execute_command(execute_ddl_file)
        self._logger.info("Stdout: %s", stdout)



"""
Usage: python -m pyexasol <command> <options>
"""

import argparse

parser = argparse.ArgumentParser(prog=f"python -m pyexasol")
subparsers = parser.add_subparsers(dest="command", metavar="<command>")
subparsers.required = True

subparsers.add_parser("script_output")
subparsers.add_parser("script_debug")

args = parser.parse_args()

if args.command == "script_output":
    raise RuntimeError(
        "'-m pyexasol script_output' is no longer available, "
        "please use '-m pyexasol_utils.script_output' instead"
    )

elif args.command == "script_debug":
    raise RuntimeError(
        "'-m pyexasol script_debug' is no longer available, "
        "please use '-m pyexasol_utils.script_output' instead"
    )



"""
This module provides `PEP-249`_ a DBAPI compliant types and type conversion definitions.
(see also `PEP-249-types`_)

.. _PEP-249-types: https://peps.python.org/pep-0249/#type-objects-and-constructors
"""

from datetime import (
    date,
    datetime,
    time,
)
from enum import Enum
from time import localtime

Date = date
Time = time
Timestamp = datetime


def DateFromTicks(ticks: int) -> date:  # pylint: disable=C0103
    """
    This function constructs an object holding a date value from the given ticks value
    (number of seconds since the epoch; see the documentation of the standard
    Python time module for details).
    """
    year, month, day = localtime(ticks)[:3]
    return Date(year, month, day)


def TimeFromTicks(ticks: int) -> time:  # pylint: disable=C0103
    """
    This function constructs an object holding a time value from the given ticks value
    (number of seconds since the epoch; see the documentation of the standard
    Python time module for details).
    """
    hour, minute, second = localtime(ticks)[3:6]
    return Time(hour, minute, second)


def TimestampFromTicks(ticks: int) -> datetime:  # pylint: disable=C0103
    """
    This function constructs an object holding a time stamp value from the
    given ticks value (number of seconds since the epoch; see the documentation
    of the standard Python time module for details).
    """
    year, month, day, hour, minute, second = localtime(ticks)[:6]
    return Timestamp(year, month, day, hour, minute, second)


class TypeCode(Enum):
    """
    Type codes for Exasol DB column types.

    See: https://github.com/exasol/websocket-api/blob/master/docs/WebsocketAPIV3.md#data-types-type-names-and-properties
    """

    Bool = "BOOLEAN"
    Char = "CHAR"
    Date = "DATE"
    Decimal = "DECIMAL"
    Double = "DOUBLE"
    Geometry = "GEOMETRY"
    IntervalDayToSecond = "INTERVAL DAY TO SECOND"
    IntervalYearToMonth = "INTERVAL YEAR TO MONTH"
    Timestamp = "TIMESTAMP"
    TimestampTz = "TIMESTAMP WITH LOCAL TIME ZONE"
    String = "VARCHAR"


class _DBAPITypeObject:
    def __init__(self, *type_codes) -> None:
        self.type_codes = type_codes

    def __eq__(self, other):
        return other in self.type_codes


STRING = _DBAPITypeObject(TypeCode.String)
# A binary type is not natively supported by Exasol
BINARY = _DBAPITypeObject(None)
NUMBER = _DBAPITypeObject(TypeCode.Decimal, TypeCode.Double)
DATETIME = _DBAPITypeObject(
    TypeCode.Date,
    TypeCode.Timestamp,
    TypeCode.TimestampTz,
    TypeCode.IntervalDayToSecond,
    TypeCode.IntervalYearToMonth,
)
# Exasol does manage indexes internally
ROWID = _DBAPITypeObject(None)



from __future__ import annotations

import hashlib
import os
import re
import socket
import socketserver
import struct
import sys
import threading
import zlib
from collections.abc import Iterable
from dataclasses import dataclass
from ssl import SSLContext
from typing import (
    TYPE_CHECKING,
    Optional,
    Union,
)

from packaging.version import Version

if TYPE_CHECKING:
    from pyexasol import ExaConnection


@dataclass
class SqlQuery:
    connection: ExaConnection
    compression: bool
    # set these values in param dictionary to ExaConnection
    column_delimiter: Optional[str] = None
    column_separator: Optional[str] = None
    columns: Optional[Iterable[str]] = None
    comment: Optional[str] = None
    csv_cols: Optional[Iterable[str]] = None
    encoding: Optional[str] = None
    format: Optional[str] = None
    null: Optional[str] = None
    row_separator: Optional[str] = None

    def _build_csv_cols(self) -> str:
        if self.csv_cols is not None:
            safe_csv_cols_regexp = re.compile(
                r"^(\d+|\d+\.\.\d+)(\sFORMAT='[^'\n]+')?$", re.IGNORECASE
            )
            for c in self.csv_cols:
                if not safe_csv_cols_regexp.match(c):
                    raise ValueError(f"Value [{c}] is not a safe csv_cols part")

            csv_cols = ",".join(self.csv_cols)
            if csv_cols != "":
                return f"({csv_cols})"

        return ""

    @staticmethod
    def _split_exa_address_into_components(exa_address: str) -> tuple[str, str | None]:
        """
        Split ip_address:port and public key from exa address, where the expected
        patterns are:
            ip_address:port
            ip_address:port/public_key
        The value for public key is expected to be a SHA-256 hash of the public key,
        which is then base64-encoded.
        """
        pattern = r"^([\d\.]+:\d+)(?:\/([a-zA-Z0-9_\-+\/]+=))?$"
        match = re.match(pattern, exa_address)
        if match is None:
            raise ValueError(
                f"Could not split exa_address {exa_address} into known components"
            )
        ip_address, public_key = match.groups()
        if not public_key:
            return ip_address, None
        return ip_address, public_key

    def _get_file_list(self, exa_address_list: list[str]) -> list[str]:
        file_ext = self._file_ext
        prefix = self._url_prefix

        csv_cols = self._build_csv_cols()
        files = []
        for i, exa_address in enumerate(exa_address_list):
            ip_address_port, public_key = self._split_exa_address_into_components(
                exa_address
            )
            statement = f"AT '{prefix}{ip_address_port}'"
            if self._requires_tls_public_key():
                if not public_key:
                    raise ValueError(
                        "Public key is required to be in the 'exa_address' for encrypted connections with Exasol DB >= 8.32.0"
                    )
                statement += f" PUBLIC KEY 'sha256//{public_key}'"
            statement += f" FILE '{str(i).rjust(3, '0')}.{file_ext}'{csv_cols}"
            files.append(statement)
        return files

    @staticmethod
    def _get_query_str(query_lines: list[Optional[str]]) -> str:
        filtered_query_lines = [q for q in query_lines if q is not None]
        return "\n".join(filtered_query_lines)

    def _requires_tls_public_key(self) -> bool:
        version = self.connection.exasol_db_version
        return (
            version is not None
            and version >= Version("8.32.0")
            and self.connection.options["encryption"]
        )

    @property
    def _column_spec(self) -> str:
        """
        Return either empty string or comma-separated list of columns in parentheses,
        e.g. '("A", "B")'
        """
        if self.columns is not None:
            formatted = [
                self.connection.format.default_format_ident(c) for c in self.columns
            ]
            comma_sep = ",".join(formatted)
            if comma_sep != "":
                return f"({comma_sep})"
        return ""

    @property
    def _column_delimiter(self) -> Optional[str]:
        if self.column_delimiter is None:
            return None
        return (
            f"COLUMN DELIMITER = {self.connection.format.quote(self.column_delimiter)}"
        )

    @property
    def _column_separator(self) -> Optional[str]:
        if self.column_separator is None:
            return None
        return (
            f"COLUMN SEPARATOR = {self.connection.format.quote(self.column_separator)}"
        )

    @property
    def _comment(self) -> Optional[str]:
        if self.comment is None:
            return None

        if "*/" in self.comment:
            raise ValueError(
                f'Invalid comment "{self.comment}". Comment must not contain "*/".'
            )
        return f"/*{self.comment}*/"

    @property
    def _encoding(self) -> Optional[str]:
        if self.encoding is None:
            return None
        return f"ENCODING = {self.connection.format.quote(self.encoding)}"

    @property
    def _file_ext(self) -> str:
        if self.format is None:
            if self.compression:
                return "gz"
            return "csv"
        if self.format not in ("gz", "bz2", "zip"):
            raise ValueError(f"Unsupported compression format: {self.format}")
        return self.format

    @property
    def _null(self) -> Optional[str]:
        if self.null is None:
            return None
        return f"NULL = {self.connection.format.quote(self.null)}"

    @property
    def _url_prefix(self) -> str:
        if self.connection.options["encryption"]:
            return "https://"
        return "http://"

    @property
    def _row_separator(self) -> Optional[str]:
        if self.row_separator is None:
            return None
        return f"ROW SEPARATOR = {self.connection.format.quote(self.row_separator)}"


@dataclass
class ImportQuery(SqlQuery):
    # set these values in param dictionary to ExaConnection
    skip: Optional[Union[str, int]] = None
    trim: Optional[str] = None

    def build_query(self, table: str, exa_address_list: list[str]) -> str:
        query_lines = [
            self._comment,
            self._get_import(table=table),
            *self._get_file_list(exa_address_list=exa_address_list),
            self._encoding,
            self._null,
            self._skip,
            self._trim,
            self._row_separator,
            self._column_separator,
            self._column_delimiter,
        ]
        return self._get_query_str(query_lines)

    @staticmethod
    def load_from_dict(
        connection: ExaConnection, compression: bool, params: dict
    ) -> ImportQuery:
        """
        Load the params dictionary into the ImportQuery class

        Keys in `params` that are not present in as attributes of the `ImportQuery`
        class will raise an Exception.
        """
        return ImportQuery(connection=connection, compression=compression, **params)

    def _get_import(self, table: str) -> str:
        return f"IMPORT INTO {table}{self._column_spec} FROM CSV"

    @property
    def _skip(self) -> Optional[str]:
        if self.skip is None:
            return None
        return f"SKIP = {self.connection.format.safe_decimal(self.skip)}"

    @property
    def _trim(self) -> Optional[str]:
        if self.trim is None:
            return None

        trim = str(self.trim).upper()
        if trim not in ("TRIM", "LTRIM", "RTRIM"):
            raise ValueError(f"Invalid value for import parameter TRIM: {trim}")
        return trim


@dataclass
class ExportQuery(SqlQuery):
    # set these values in param dictionary to ExaConnection
    delimit: Optional[str] = None
    with_column_names: Optional[str] = None

    def build_query(self, table: str, exa_address_list: list[str]) -> str:
        query_lines = [
            self._comment,
            self._get_export(table=table),
            *self._get_file_list(exa_address_list=exa_address_list),
            self._delimit,
            self._encoding,
            self._null,
            self._row_separator,
            self._column_separator,
            self._column_delimiter,
            self._with_column_names,
        ]
        return self._get_query_str(query_lines)

    @staticmethod
    def load_from_dict(
        connection: ExaConnection, compression: bool, params: dict
    ) -> ExportQuery:
        """
        Load the params dictionary into the ExportQuery class

        Keys in `params` that are not present in as attributes of the `ExportQuery`
        class will raise an Exception.
        """
        return ExportQuery(connection=connection, compression=compression, **params)

    def _get_export(self, table: str) -> str:
        return f"EXPORT {table}{self._column_spec} INTO CSV"

    @property
    def _delimit(self) -> Optional[str]:
        if self.delimit is None:
            return None

        delimit = str(self.delimit).upper()
        if delimit not in ("AUTO", "ALWAYS", "NEVER"):
            raise ValueError(f"Invalid value for export parameter DELIMIT: {delimit}")
        return f"DELIMIT={delimit}"

    @property
    def _with_column_names(self) -> Optional[str]:
        if self.with_column_names is None:
            return None
        return "WITH COLUMN NAMES"


class ExaSQLThread(threading.Thread):
    """
    Thread class which re-throws any Exception to parent thread
    """

    def __init__(self, connection: ExaConnection, compression: bool):
        self.connection = connection
        self.compression = compression

        self.params: dict = {}
        self.http_thread = None
        self.exa_address_list: list[str] = []
        self.exc = None

        super().__init__()

    def set_http_thread(self, http_thread):
        self.http_thread = http_thread
        self.exa_address_list = [http_thread.exa_address]

    def set_exa_address_list(self, exa_address_list):
        self.exa_address_list = exa_address_list

    def run(self):
        try:
            self.run_sql()
        except BaseException as e:
            self.exc = e

            # In case of SQL error stop HTTP server, close pipes and interrupt I/O in callback function
            if self.http_thread:
                self.http_thread.terminate()

    def run_sql(self):
        pass

    def join_with_exc(self, *args):
        super().join(*args)

        if self.exc:
            raise self.exc


class ExaSQLExportThread(ExaSQLThread):
    """
    Build and run EXPORT query into separate thread
    Main thread is busy outputting data in callbacks
    """

    def __init__(
        self,
        connection: ExaConnection,
        compression: bool,
        query_or_table,
        export_params: dict,
    ):
        super().__init__(connection, compression)

        self.query_or_table = query_or_table
        self.params = export_params

    def run_sql(self):
        if (
            isinstance(self.query_or_table, tuple)
            or str(self.query_or_table).strip().find(" ") == -1
        ):
            export_table = self.connection.format.default_format_ident(
                self.query_or_table
            )
        else:
            # New lines are mandatory to handle queries with single-line comments '--'
            export_query = self.query_or_table.lstrip(" \n").rstrip(" \n;")
            export_table = f"(\n{export_query}\n)"

            if self.params.get("columns"):
                raise ValueError(
                    "Export option 'columns' is not compatible with SQL query export source"
                )

        export_query = ExportQuery.load_from_dict(
            connection=self.connection, compression=self.compression, params=self.params
        ).build_query(table=export_table, exa_address_list=self.exa_address_list)
        self.connection.execute(export_query)


class ExaSQLImportThread(ExaSQLThread):
    """
    Build and run EXPORT query into separate thread
    Main thread is busy parsing results in callbacks
    """

    def __init__(
        self,
        connection: ExaConnection,
        compression: bool,
        table: str,
        import_params: dict,
    ):
        super().__init__(connection, compression)

        self.table = table
        self.params = import_params

    def run_sql(self):
        table = self.connection.format.default_format_ident(self.table)

        import_query = ImportQuery.load_from_dict(
            connection=self.connection, compression=self.compression, params=self.params
        ).build_query(table=table, exa_address_list=self.exa_address_list)
        self.connection.execute(import_query)


class ExaHttpThread(threading.Thread):
    """
    HTTP communication and compression / decompression is offloaded to a separate thread.
    PyExasol uses a thread instead of a subprocess or multiprocessing to avoid
    compatibility issues on Windows operating systems. For further details, see
    - https://github.com/exasol/pyexasol/issues/73
    - https://pythonforthelab.com/blog/differences-between-multiprocessing-windows-and-linux/
    """

    def __init__(self, ipaddr: str, port: int, compression: bool, encryption: bool):
        self.server = ExaTCPServer(
            (ipaddr, port),
            ExaHttpRequestHandler,
            compression=compression,
            encryption=encryption,
        )

        self.read_pipe = self.server.read_pipe
        self.write_pipe = self.server.write_pipe

        self.exc = None

        super().__init__()

    @property
    def exa_address(self) -> str:
        address = f"{self.server.exa_address_ipaddr}:{self.server.exa_address_port}"
        if public_key := self.server.exa_address_public_key:
            address = f"{address}/{public_key}"
        return address

    def run(self):
        try:
            # Handle exactly one HTTP request
            # Exit loop if thread was explicitly terminated prior to receiving HTTP request
            while self.server.total_clients == 0 and not self.server.is_terminated:
                self.server.handle_request()
        except BaseException as e:
            self.exc = e
        finally:
            self.server.server_close()

    def join(self, timeout=None):
        self.server.can_finish_get.set()
        super().join(timeout)

    def join_with_exc(self):
        self.join()

        if self.exc:
            raise self.exc

    def terminate(self):
        self.server.is_terminated = True
        self.server.can_finish_get.set()

        # Must close pipes here to prevent infinite lock in callback function
        # Termination pipe order is important for Windows
        self.write_pipe.close()
        self.read_pipe.close()


class ExaHTTPTransportWrapper:
    """
    Wrapper for :ref:`http_transport_parallel`.

    You may create this wrapper using :func:`pyexasol.http_transport`.

    Note:

        Starts an HTTP server, obtains the address (the ``"ipaddr:port"`` string),
        and sends it to the parent process.

        Block into ``export_*()`` or ``import_*()`` call,
        wait for incoming connection, process data and exit.
    """

    def __init__(
        self,
        ipaddr: str,
        port: int,
        compression: bool = False,
        encryption: bool = True,
    ):
        self.http_thread = ExaHttpThread(ipaddr, port, compression, encryption)
        self.http_thread.start()

    @property
    def exa_address(self) -> str:
        """
        Internal Exasol address as ``ipaddr:port`` string.

        Note:
            This string should be passed from child processes to parent process
            and used as an argument for ``export_parallel()`` and
            ``import_parallel()`` functions.
        """
        return self.http_thread.exa_address

    def get_proxy(self):
        """
        Caution:
            **DEPRECATED**, please use ``.exa_address`` property
        """
        return self.http_thread.exa_address

    def export_to_callback(self, callback, dst, callback_params=None):
        """
        Exports chunk of data using callback function.

        Args:
            callback:
                Callback function.
            dst:
                Export destination for callback function.
            callback_params:
                Dict with additional parameters for callback function.

        Returns:
            Result of the callback function.

        Note:
            You may use exactly the same callbacks utilized by standard
            non-parallel ``export_to_callback()`` function.
        """
        if not callable(callback):
            raise ValueError("Callback argument is not callable")

        if callback_params is None:
            callback_params = {}

        try:
            result = callback(self.http_thread.read_pipe, dst, **callback_params)

            self.http_thread.read_pipe.close()
            self.http_thread.join_with_exc()

            return result

        except (Exception, KeyboardInterrupt) as e:
            self.http_thread.terminate()
            self.http_thread.join()

            raise e

    def import_from_callback(self, callback, src, callback_params=None):
        """
        Import chunk of data using callback function.

        Args:
            callback:
                Callback function.
            src:
                Import source for the callback function.
            callback_params:
                Dict with additional parameters for the callback function.

        Returns:
            Result of callback function

        Note:
            You may use exactly the same callbacks utilized by standard
            non-parallel ``import_from_callback()`` function.
        """
        if not callable(callback):
            raise ValueError("Callback argument is not callable")

        if callback_params is None:
            callback_params = {}

        try:
            result = callback(self.http_thread.write_pipe, src, **callback_params)

            self.http_thread.write_pipe.close()
            self.http_thread.join_with_exc()

            return result

        except (Exception, KeyboardInterrupt) as e:
            self.http_thread.terminate()
            self.http_thread.join()

            raise e

    def __repr__(self):
        return f"<{self.__class__.__name__} exa_address={self.exa_address}>"


class ExaTCPServer(socketserver.TCPServer):
    exa_address_ipaddr: str
    exa_address_port: int
    exa_address_public_key: Optional[str] = None

    total_clients: int = 0
    is_terminated: bool = False

    timeout: int | None = 1

    def __init__(self, *args, **kwargs):
        self.compression: bool = kwargs.pop("compression", False)
        self.encryption: bool = kwargs.pop("encryption", True)

        r_fd, w_fd = os.pipe()

        self.read_pipe = open(r_fd, "rb", 0)
        self.write_pipe = open(w_fd, "wb", 0)

        # GET method calls (IMPORT) require extra protection
        #
        # Callback function may close pipe abruptly and raise an exception
        # It may cause partial valid data to be sent to Exasol server
        #
        # This event waits for callback function to finish before sending final chunk of data
        #
        # If callback function failed with exception, HTTP thread will be terminated
        # Final chunk will not be sent, causing IMPORT query to fail and to discard partial data
        self.can_finish_get = threading.Event()

        super().__init__(*args, **kwargs)

    def server_bind(self):
        self.set_sock_opts()

        """ Special Exasol packet to establish tunneling and return an internal Exasol address, which can be used in a query """
        self.socket.connect(self.server_address)
        self.socket.sendall(struct.pack("iii", 0x02212102, 1, 1))
        _, port, ipaddr = struct.unpack("ii16s", self.socket.recv(24))

        self.exa_address_ipaddr = ipaddr.replace(b"\x00", b"").decode()
        self.exa_address_port = port

        if self.encryption:
            context, public_key_sha = self.generate_adhoc_ssl_context()
            self.socket = context.wrap_socket(
                self.socket, server_side=True, do_handshake_on_connect=False
            )
            self.exa_address_public_key = public_key_sha

    def server_activate(self):
        pass

    def get_request(self):
        return self.socket, self.server_address

    def shutdown_request(self, request):
        pass

    def close_request(self, request):
        pass

    def set_sock_opts(self):
        # only large packets are expected for HTTP transport
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        # default keep-alive once a minute, 5 probes
        keepidle = 60
        keepintvl = 60
        keepcnt = 5

        if sys.platform.startswith("linux"):
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, keepidle)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, keepintvl)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, keepcnt)

        elif sys.platform.startswith("darwin"):
            # TCP_KEEPALIVE = 0x10
            # https://bugs.python.org/issue34932
            self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.socket.setsockopt(socket.IPPROTO_TCP, 0x10, keepidle)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, keepintvl)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, keepcnt)

        elif sys.platform.startswith("win"):
            self.socket.ioctl(
                socket.SIO_KEEPALIVE_VALS, (1, keepidle * 1000, keepintvl * 1000)
            )

    @staticmethod
    def generate_adhoc_ssl_context() -> tuple[SSLContext, str]:
        """
        Create temporary self-signed certificate for encrypted HTTP transport
        Exasol does not check validity of certificates
        """
        from base64 import b64encode
        from datetime import (
            datetime,
            timedelta,
            timezone,
        )
        from pathlib import Path
        from ssl import (
            CERT_NONE,
            PROTOCOL_TLS_SERVER,
        )
        from tempfile import TemporaryDirectory

        from cryptography import x509
        from cryptography.hazmat.primitives import (
            hashes,
            serialization,
        )
        from cryptography.hazmat.primitives.asymmetric import rsa
        from cryptography.x509.oid import NameOID

        key_pair = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )

        # For a self-signed certificate, subject and issuer are identical.
        subject = issuer = x509.Name(
            [
                x509.NameAttribute(NameOID.COUNTRY_NAME, "DE"),
                x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "Franconia"),
                x509.NameAttribute(NameOID.LOCALITY_NAME, "Nuremberg"),
                x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Exasol AG"),
                x509.NameAttribute(NameOID.COMMON_NAME, "exasol.com"),
            ]
        )
        today = datetime.now(timezone.utc)
        cert = (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(key_pair.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(today)
            .not_valid_after(today + timedelta(days=365))
            .sign(key_pair, hashes.SHA256())
        )

        der_data = key_pair.public_key().public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        sha256_hash = hashlib.sha256(der_data).digest()
        base64_encoded = b64encode(sha256_hash)
        public_key_sha256 = base64_encoded.decode("utf-8")

        # TemporaryDirectory is used instead of NamedTemporaryFile for compatibility with Windows
        with TemporaryDirectory(prefix="pyexasol_ssl_") as tempdir:
            directory = Path(tempdir)

            cert_file = open(directory / "cert", "wb")
            cert_file.write(cert.public_bytes(serialization.Encoding.PEM))
            cert_file.close()

            key_file = open(directory / "key", "wb")
            key_file.write(
                key_pair.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            )
            key_file.close()

            context = SSLContext(PROTOCOL_TLS_SERVER)
            context.verify_mode = CERT_NONE
            context.load_cert_chain(certfile=cert_file.name, keyfile=key_file.name)

            return context, public_key_sha256


class ExaHttpRequestHandler(socketserver.StreamRequestHandler):
    success_headers = (
        b"HTTP/1.1 200 OK\r\n"
        b"Connection: close\r\n"
        b"Transfer-Encoding: chunked\r\n"
        b"\r\n"
    )

    error_headers = (
        b"HTTP/1.1 500 Internal Server Error\r\n" b"Connection: close\r\n" b"\r\n"
    )

    server: ExaTCPServer

    def handle(self):
        self.server.total_clients += 1

        # Extract method from the first header
        method = str(self.rfile.readline(), "iso-8859-1").split()[0]

        # Skip all other headers
        while self.rfile.readline() != b"\r\n":
            pass

        if method == "PUT":
            if self.server.compression:
                self.method_put_compressed()
            else:
                self.method_put_raw()
        elif method == "GET":
            if self.server.compression:
                self.method_get_compressed()
            else:
                self.method_get_raw()
        else:
            raise RuntimeError(f"Unsupported HTTP method [{method}]")

    def method_put_raw(self):
        try:
            while not self.server.is_terminated:
                data = self.read_chunk()

                if data is None:
                    break

                self.server.write_pipe.write(data)

        except Exception as e:
            self.write_error_headers()
            raise e

        else:
            self.write_success_headers()
            self.write_final_chunk()

        finally:
            self.server.write_pipe.close()

    def method_put_compressed(self):
        try:
            d = zlib.decompressobj(wbits=16 + zlib.MAX_WBITS)

            while not self.server.is_terminated:
                data = self.read_chunk()

                if data is None:
                    self.server.write_pipe.write(d.flush())
                    break

                self.server.write_pipe.write(d.decompress(data))

        except Exception as e:
            self.write_error_headers()
            raise e

        else:
            self.write_success_headers()
            self.write_final_chunk()

        finally:
            self.server.write_pipe.close()

    def method_get_raw(self):
        try:
            self.write_success_headers()

            while not self.server.is_terminated:
                # Exasol server produces chunks of this size (without chunk header part)
                data = self.server.read_pipe.read(65524)

                if data is None or len(data) == 0:
                    break

                self.write_chunk(data)

        except Exception as e:
            raise e

        finally:
            self.server.read_pipe.close()

        if self.server.can_finish_get.wait() and not self.server.is_terminated:
            self.write_final_chunk()

    def method_get_compressed(self):
        try:
            self.write_success_headers()
            c = zlib.compressobj(wbits=16 + zlib.MAX_WBITS)

            while not self.server.is_terminated:
                #  Linux common pipe buffer, 64Kb
                data = self.server.read_pipe.read(65536)

                if data is None or len(data) == 0:
                    self.write_chunk(c.flush(zlib.Z_FINISH))
                    break

                self.write_chunk(c.compress(data))

        except Exception as e:
            raise e

        finally:
            self.server.read_pipe.close()

        if self.server.can_finish_get.wait() and not self.server.is_terminated:
            self.write_final_chunk()

    def read_chunk(self):
        hex_length = self.rfile.readline().rstrip()

        if len(hex_length) == 0:
            chunk_len = 0
        else:
            chunk_len = int(hex_length, 16)

        if chunk_len == 0:
            return None

        data = self.rfile.read(chunk_len)

        if self.rfile.read(2) != b"\r\n":
            raise RuntimeError("Invalid chunk delimiter in HTTP stream")

        return data

    def write_chunk(self, data):
        chunk_len = len(data)

        if chunk_len == 0:
            return

        self.wfile.write(b"%X\r\n%b\r\n" % (chunk_len, data))

    def write_final_chunk(self):
        self.wfile.write(b"0\r\n\r\n")

    def write_success_headers(self):
        self.wfile.write(self.success_headers)

    def write_error_headers(self):
        self.wfile.write(self.error_headers)
